# -*- coding: utf-8 -*-
# clean
""" Generic heating controller. """

from dataclasses import dataclass

# Owned
from typing import List, Optional, Any

from dataclasses_json import dataclass_json

from hisim import utils
from hisim import component as cp
from hisim.components import generic_hot_water_storage_modular
from hisim.components.building import Building
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters

__authors__ = "edited Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class L2GenericHeatConfig(cp.ConfigBase):
    """L2 Controller Config."""

    building_name: str
    name: str
    source_weight: int
    t_min_heating_in_celsius: float
    t_max_heating_in_celsius: float
    cooling_considered: bool
    t_min_cooling_in_celsius: Optional[float]
    t_max_cooling_in_celsius: Optional[float]
    day_of_heating_season_begin: Optional[int]
    day_of_heating_season_end: Optional[int]

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return L2GenericHeatController.get_full_classname()

    @staticmethod
    def get_default_config_heating(
        name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Default config for the heating controller."""
        config = L2GenericHeatConfig(
            building_name=building_name,
            name="L2HeatingTemperatureController_" + name,
            source_weight=1,
            t_min_heating_in_celsius=20.0,
            t_max_heating_in_celsius=22.0,
            cooling_considered=False,
            t_min_cooling_in_celsius=23,
            t_max_cooling_in_celsius=25,
            day_of_heating_season_begin=270,
            day_of_heating_season_end=150,
        )
        return config

    @staticmethod
    def get_default_config_buffer_heating(
        name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Default Config for the buffer temperature."""
        config = L2GenericHeatConfig(
            building_name=building_name,
            name="L2BufferTemperatureController_" + name,
            source_weight=1,
            t_min_heating_in_celsius=30.0,
            t_max_heating_in_celsius=50.0,
            cooling_considered=False,
            t_min_cooling_in_celsius=23,
            t_max_cooling_in_celsius=25,
            day_of_heating_season_begin=270,
            day_of_heating_season_end=150,
        )
        return config

    @staticmethod
    def get_default_config_waterheating(
        name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Generate Default Config for a DHW controller."""
        config = L2GenericHeatConfig(
            building_name=building_name,
            name="L2DHWTemperatureController_" + name,
            source_weight=1,
            t_min_heating_in_celsius=50.0,
            t_max_heating_in_celsius=80.0,
            cooling_considered=False,
            t_min_cooling_in_celsius=None,
            t_max_cooling_in_celsius=None,
            day_of_heating_season_begin=None,
            day_of_heating_season_end=None,
        )
        return config


class L2GenericHeatControllerState:
    """Data class that saves the state of the controller."""

    def __init__(
        self,
        timestep_actual: int = -1,
        state: int = 0,
        compulsory: int = 0,
        count: int = 0,
    ):
        """Initializes the class."""
        self.timestep_actual: int = timestep_actual
        self.state: int = state
        self.compulsory: int = compulsory
        self.count: int = count

    def clone(self):
        """Clones itself."""
        return L2GenericHeatControllerState(
            timestep_actual=self.timestep_actual,
            state=self.state,
            compulsory=self.compulsory,
            count=self.count,
        )

    def is_first_iteration(self, timestep):
        """Only called for first iteration."""
        if self.timestep_actual + 1 == timestep:
            self.timestep_actual += 1
            self.compulsory = 0
            self.count = 0
            return True
        return False

    def is_compulsory(self):
        """Returns compulsory value."""
        if self.count <= 1:
            self.compulsory = 0
        else:
            self.compulsory = 1

    def activate(self):
        """Ativates."""
        self.state = 1
        self.compulsory = 1
        self.count += 1

    def deactivate(self):
        """Deactivates the control signal."""
        self.state = 0
        self.compulsory = 1
        self.count += 1


class L2GenericHeatController(cp.Component):
    """L2 heat pump controller. Processes signals ensuring comfort temperature of building.

    Gets available surplus electricity and the temperature of the storage or building to control as input,
    and outputs control signal 0/1 for turn off/switch on based on comfort temperature limits and available electricity.
    It optionally has different modes for cooling and heating selected by the time of the year.

    Parameters
    ----------
    source_weight : int, optional
        Weight of component, relevant if there is more than one component of same type, defines hierachy in control. The default is 1.
    T_min_heating: float, optional
        Minimum comfortable temperature for residents during heating period, in °C. The default is 19 °C.
    T_max_heating: float, optional
        Maximum comfortable temperature for residents during heating period, in °C. The default is 23 °C.
    T_tolerance : float, optional
        Temperature difference the building may go below or exceed the comfort temperature band with, because of recommendations from L3. The default is 1 °C.

    """

    # Inputs
    ReferenceTemperature = "ReferenceTemperature"

    # Outputs
    l2_device_signal = "l2_DeviceSignal"

    # #Forecasts
    # HeatPumpLoadForecast = "HeatPumpLoadForecast"

    # Similar components to connect to:
    # 1. Building
    # 2. HeatPump

    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: L2GenericHeatConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """For initializing."""
        if not config.__class__.__name__ == L2GenericHeatConfig.__name__:
            raise ValueError("Wrong config class.")
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.config: L2GenericHeatConfig = config

        """ Initializes the class. """
        self.source_weight: int = config.source_weight
        self.cooling_considered: bool = config.cooling_considered
        if self.cooling_considered:
            if config.day_of_heating_season_begin is None:
                raise ValueError("Day of heating season begin was None")
            if config.day_of_heating_season_end is None:
                raise ValueError("Day of heating season end was None")
            self.heating_season_begin = (
                config.day_of_heating_season_begin * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            )
            self.heating_season_end = (
                config.day_of_heating_season_end * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            )
        self.state: L2GenericHeatControllerState = L2GenericHeatControllerState()
        self.previous_state: L2GenericHeatControllerState = L2GenericHeatControllerState()

        # Component Outputs
        self.l2_device_signal_channel: cp.ComponentOutput = self.add_output(
            self.component_name, self.l2_device_signal, LoadTypes.ON_OFF, Units.BINARY
        )

        # Component Inputs
        self.reference_temperature_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ReferenceTemperature,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            mandatory=True,
        )

        self.add_default_connections(self.get_default_connections_from_buildings())
        self.add_default_connections(self.get_default_connections_from_generic_hot_water_storage_modular())

    def get_default_connections_from_buildings(self):
        """Sets the default connections for the building."""

        connections = []
        building_classname = Building.get_classname()
        connections.append(
            cp.ComponentConnection(
                L2GenericHeatController.ReferenceTemperature,
                building_classname,
                Building.TemperatureMean,
            )
        )
        return connections

    def get_default_connections_from_generic_hot_water_storage_modular(self):
        """Sets default connections for the boiler."""

        connections = []
        hotwaterstorage_classname = generic_hot_water_storage_modular.HotWaterStorage.get_classname()
        connections.append(
            cp.ComponentConnection(
                L2GenericHeatController.ReferenceTemperature,
                hotwaterstorage_classname,
                generic_hot_water_storage_modular.HotWaterStorage.TemperatureMean,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def control_cooling(
        self,
        t_control: float,
        t_min_cooling: Optional[float],
        t_max_cooling: Optional[float],
    ) -> None:
        """Controls the cooling."""
        if t_min_cooling is None:
            raise ValueError("T_min_cooling was None.")
        if t_max_cooling is None:
            raise ValueError("T_max_cooling was None.")
        """ Controls the cooling. """
        if t_control > t_max_cooling:
            # start cooling if temperature exceeds upper limit
            self.state.activate()
            self.previous_state.activate()
        elif t_control < t_min_cooling:
            # stop cooling if temperature goes below lower limit
            self.state.deactivate()
            self.previous_state.deactivate()
        else:
            if self.state.compulsory == 1:
                # use previous state if it is compulsory
                pass
            else:
                # use previous state if l3 was not available
                self.state = self.previous_state.clone()

    def control_heating(self, t_control: float, t_min_heating: float, t_max_heating: float) -> None:
        """Controles the heating."""
        if t_control > t_max_heating:
            # stop heating if temperature exceeds upper limit
            self.state.deactivate()
            self.previous_state.deactivate()

        elif t_control < t_min_heating:
            # start heating if temperature goes below lower limit
            self.state.activate()
            self.previous_state.activate()
        else:
            if self.state.compulsory == 1:
                # use previous state if it compulsory
                pass
            else:
                # use previous state if temperature is in given limit
                self.state = self.previous_state.clone()

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores previous state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """For double checking results."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Core Simulation function."""
        if force_convergence:
            return
        # check demand, and change state of self.has_heating_demand, and self._has_cooling_demand
        t_control = stsv.get_input_value(self.reference_temperature_channel)

        # check if it is the first iteration and reset compulsory and timestep_of_last_activation in state and previous_state
        if self.state.is_first_iteration(timestep):
            self.previous_state.is_first_iteration(timestep)
        if self.cooling_considered:
            # check out during cooling season
            if self.heating_season_begin > timestep > self.heating_season_end:
                self.control_cooling(
                    t_control=t_control,
                    t_min_cooling=self.config.t_min_cooling_in_celsius,
                    t_max_cooling=self.config.t_max_cooling_in_celsius,
                )
            # check out during heating season
            else:
                self.control_heating(
                    t_control=t_control,
                    t_min_heating=self.config.t_min_heating_in_celsius,
                    t_max_heating=self.config.t_max_heating_in_celsius,
                )

        # check out during heating season
        else:
            self.control_heating(
                t_control=t_control,
                t_min_heating=self.config.t_min_heating_in_celsius,
                t_max_heating=self.config.t_max_heating_in_celsius,
            )
        stsv.set_output_value(self.l2_device_signal_channel, self.state.state)

    def write_to_report(self) -> List[str]:
        """Writes the information of the current component to the report."""
        lines: List[str] = []
        lines.append(f"Name: {self.component_name + str(self.config.source_weight)}")
        lines.append(f"upper set temperature: {self.config.t_max_heating_in_celsius:4.0f} °C")
        lines.append(f"lower set temperature: {self.config.t_min_heating_in_celsius:4.0f} °C")
        return lines



"""Generic heat water storage module."""

# clean

# Generic/Built-in
from typing import Any, List
from dataclasses import dataclass
from dataclasses_json import dataclass_json

# Owned
from hisim.component import Component, ComponentInput, ComponentOutput, ConfigBase
from hisim import component as cp
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters


__authors__ = "Maximilian Hillen"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Maximilian Hillen"
__email__ = "maximilian.hillen@rwth-aachen.de"
__status__ = ""


@dataclass_json
@dataclass
class HeatStorageConfig(ConfigBase):
    """Heat Storage Config class."""

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return HeatStorage.get_full_classname()

    building_name: str
    name: str
    volume_sp_heating_water: float
    volume_sp_warm_water: float
    temperature_of_warm_water_extraction: float
    ambient_temperature: float
    temperature_sp_ww: float
    temperature_sp_hw: float

    @classmethod
    def get_default_heat_storage_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Get default config."""
        config = HeatStorageConfig(
            building_name=building_name,
            name="HeatStorage",
            volume_sp_heating_water=1000,
            volume_sp_warm_water=100,
            temperature_of_warm_water_extraction=32,
            ambient_temperature=15,
            temperature_sp_ww=40,
            temperature_sp_hw=40,
        )
        return config


@dataclass_json
@dataclass
class HeatStorageControllerConfig(ConfigBase):
    """Heat Storage Controller Config."""

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return HeatStorageController.get_full_classname()

    building_name: str
    name: str
    initial_temperature_building: float
    initial_temperature_heating_storage: float
    heating_load_of_building_in_watt: float

    @classmethod
    def get_default_heat_storage_controller_config(
        cls,
        heating_load_of_building_in_watt: float,
        building_name: str = "BUI1",
    ) -> Any:
        """Get default config."""
        config = HeatStorageControllerConfig(
            building_name=building_name,
            name="HeatStorageController",
            initial_temperature_building=20,
            initial_temperature_heating_storage=35,
            heating_load_of_building_in_watt=heating_load_of_building_in_watt,
        )
        return config


class HeatStorageState:
    """Heat Storage State class."""

    def __init__(self, temperature_sp_ww: float, temperature_sp_hw: float) -> None:
        """Intitialize the class."""
        self.temperature_sp_ww = temperature_sp_ww
        self.temperature_sp_hw = temperature_sp_hw

    def clone(self) -> Any:
        """Clones the state."""
        return HeatStorageState(
            temperature_sp_ww=self.temperature_sp_ww,
            temperature_sp_hw=self.temperature_sp_hw,
        )


class HeatStorage(Component):
    """Heat Storage class.

    This is a combined storage: buffer storage for heating, and hot water storage for hot water demand.
    It needs, hot water demand, heating demand, building temperature and a control signal choosing which of the two storages to heat as inputs.
    In addition it relies on Outside Temperature, and ThermalInputs of up to 5 heat sources.
    Based on this it evaluates the temperature in the storages based on the energy balance.
    """

    # Inputs
    ThermalDemandHeatingWater = "ThermalDemandHeatingWater"  # Heating Water to regulate room Temperature
    ThermalDemandWarmWater = "ThermalDemandHeating"  # Warmwater for showering, washing etc...
    ControlSignalChooseStorage = "ControlSignalChooseStorage"
    BuildingTemperature = "BuildingTemperature"

    OutsideTemperature = "OutsideTemperature"
    ThermalInputPower1 = "ThermalInputPower1"
    ThermalInputPower2 = "ThermalInputPower2"
    ThermalInputPower3 = "ThermalInputPower3"
    ThermalInputPower4 = "ThermalInputPower4"
    ThermalInputPower5 = "ThermalInputPower5"

    # Outputs
    WaterOutputTemperatureHeatingWater = "WaterOutputTemperatureHeatingWater"
    WaterOutputTemperatureWarmWater = "WaterOutputTemperatureWarmWater"
    WaterOutputStorageforHeaters = "WaterOutputStorageforHeaters"
    # StorageWarmWaterTemperature="StorageWarmWaterTemperature"
    StorageEnergyLoss = "StorageEnergyLoss"
    RealHeatForBuilding = "RealHeatForBuilding"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: HeatStorageConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.heat_storage_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.volume_sp_heating_water = self.heat_storage_config.volume_sp_heating_water
        self.volume_sp_warm_water = self.heat_storage_config.volume_sp_warm_water
        self.temperature_of_warm_water_extratcion = self.heat_storage_config.temperature_of_warm_water_extraction
        self.ambient_temperature = self.heat_storage_config.ambient_temperature
        self.temperature_sp_warm_water = self.heat_storage_config.temperature_sp_ww
        self.temperature_sp_heating_water = self.heat_storage_config.temperature_sp_hw
        self.specific_heat_capacity_water = 4812

        self.state = HeatStorageState(
            self.heat_storage_config.temperature_sp_ww,
            self.heat_storage_config.temperature_sp_hw,
        )
        self.previous_state = self.state.clone()

        self.thermal_demand_heating_water_channel: ComponentInput = self.add_input(
            self.component_name,
            self.ThermalDemandHeatingWater,
            lt.LoadTypes.WARM_WATER,
            lt.Units.WATT,
            False,
        )

        self.thermal_demand_warm_water_channel: ComponentInput = self.add_input(
            self.component_name,
            self.ThermalDemandWarmWater,
            lt.LoadTypes.WARM_WATER,
            lt.Units.WATT,
            False,
        )
        self.control_signal_choose_storage_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ControlSignalChooseStorage,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            False,
        )
        self.building_temperature_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.BuildingTemperature,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            False,
        )
        self.thermal_input_power_one_channel: ComponentInput = self.add_input(
            self.component_name,
            self.ThermalInputPower1,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            False,
        )
        self.thermal_input_power_two_channel: ComponentInput = self.add_input(
            self.component_name,
            self.ThermalInputPower2,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            False,
        )
        self.thermal_input_power_three_channel: ComponentInput = self.add_input(
            self.component_name,
            self.ThermalInputPower3,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            False,
        )
        self.thermal_input_power_four_channel: ComponentInput = self.add_input(
            self.component_name,
            self.ThermalInputPower4,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            False,
        )
        self.thermal_input_power_five_channel: ComponentInput = self.add_input(
            self.component_name,
            self.ThermalInputPower5,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            False,
        )

        self.water_output_temperature_heating_water_channel: ComponentOutput = self.add_output(
            self.component_name,
            self.WaterOutputTemperatureHeatingWater,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            output_description=f"here a description for {self.WaterOutputTemperatureHeatingWater} will follow.",
        )
        self.water_output_temperature_warm_water_channel: ComponentOutput = self.add_output(
            self.component_name,
            self.WaterOutputTemperatureWarmWater,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            output_description=f"here a description for {self.WaterOutputTemperatureWarmWater} will follow.",
        )
        self.storage_energy_loss_channel: ComponentOutput = self.add_output(
            self.component_name,
            self.StorageEnergyLoss,
            lt.LoadTypes.ANY,
            lt.Units.WATT,
            output_description=f"here a description for {self.StorageEnergyLoss} will follow.",
        )
        self.water_output_temperature_storage_for_heaters_channel: ComponentOutput = self.add_output(
            self.component_name,
            self.WaterOutputStorageforHeaters,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            output_description=f"here a description for {self.WaterOutputStorageforHeaters} will follow.",
        )
        self.real_heat_for_building_channel: ComponentOutput = self.add_output(
            self.component_name,
            self.RealHeatForBuilding,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            output_description=f"here a description for {self.RealHeatForBuilding} will follow.",
        )

    def write_to_report(self) -> List[str]:
        """Write to report."""
        lines = []
        lines.append("Name: HeatWaterStorage")
        lines.append(f"Volume Warm Water Storage [L]: {self.volume_sp_warm_water}")
        lines.append(f"Volume Heat Water Storage [L]: {self.volume_sp_heating_water}")
        lines.append(f"Temperature of Warm Water Extraction [°C]: {self.temperature_of_warm_water_extratcion}")
        lines.append(f"Ambient Temperature [°C]: {self.ambient_temperature}")
        lines.append(f"Temperature Warm Water Storage [°C]: {self.temperature_sp_warm_water}")
        lines.append(f"temperature Heat Water Storage [°C]: {self.temperature_sp_heating_water}")

        return self.heat_storage_config.get_string_dict() + lines

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def adding_all_possible_mass_flows(self, stsv: cp.SingleTimeStepValues) -> Any:
        """Adding all possible mass flows."""
        production: float = 0
        # function to add all possible mass flows

        if self.thermal_input_power_one_channel.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power_one_channel) + production

        if self.thermal_input_power_two_channel.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power_two_channel) + production

        if self.thermal_input_power_three_channel.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power_three_channel) + production

        if self.thermal_input_power_four_channel.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power_four_channel) + production

        if self.thermal_input_power_five_channel.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power_five_channel) + production

        return production

    def calculate_new_storage_temperature(
        self,
        seconds_per_timestep: int,
        temperature_sp: float,
        production: float,
        last: float,
        c_w: float,
        volume_sp: float,
    ) -> Any:
        """Calculate new storage temperature."""

        temperature_ext_sp = self.ambient_temperature

        mass_sp_h = volume_sp * 0.99  # Vereinfachung
        heat_loss_storage = 0.0038 * volume_sp + 0.85  # Heatloss Storage

        # calcutae new Storage Temp.
        temperature_sp = (
            temperature_sp
            + (1 / (mass_sp_h * c_w))
            * (production - last - heat_loss_storage * (temperature_sp - temperature_ext_sp))
            * seconds_per_timestep
        )
        # T_SP = T_sp + (dt/(m_SP_h*c_w))*(P_h_HS*(T_sp-T_ext_SP) - last*(T_sp-T_ext_SP) - UA_SP*(T_sp-T_ext_SP))
        # Correction Calculation
        # T_sp_k = (T_sp+T_SP)/2
        # T_vl = T_sp_k+2.5

        # calcutae new Storage Temp.
        # T_SP = T_sp + (1/(m_SP_h*c_w))*( production- last - UA_SP*(T_sp_k-T_ext_SP))*dt

        return temperature_sp, heat_loss_storage

    # def regarding_heating_water_storage (self, T_sp: int):

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""

        temperature_sp_var_ww = self.state.temperature_sp_ww  # Start-Temp-Storage
        temperature_sp_var_hw = self.state.temperature_sp_hw  # Start-Temp-Storage

        last_var_ww = stsv.get_input_value(self.thermal_demand_warm_water_channel)
        last_var_hw = stsv.get_input_value(self.thermal_demand_heating_water_channel)

        result_ww = [temperature_sp_var_ww, 0]
        result_hw = [temperature_sp_var_hw, 0]
        temperature_sp_c = (temperature_sp_var_ww + temperature_sp_var_hw) / 2

        if (
            stsv.get_input_value(self.control_signal_choose_storage_channel) == 1
        ):  # choose to heat up warm water storage
            production_var = self.adding_all_possible_mass_flows(stsv)
            result_ww = self.calculate_new_storage_temperature(
                seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep,
                temperature_sp=temperature_sp_var_ww,
                production=production_var,
                last=last_var_ww,
                c_w=self.specific_heat_capacity_water,
                volume_sp=self.volume_sp_warm_water,
            )
            temperature_sp_c = result_ww[0]
            production_var = 0
            result_hw = self.calculate_new_storage_temperature(
                seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep,
                temperature_sp=temperature_sp_var_hw,
                production=production_var,
                last=last_var_hw,
                c_w=self.specific_heat_capacity_water,
                volume_sp=self.volume_sp_heating_water,
            )

        elif (
            stsv.get_input_value(self.control_signal_choose_storage_channel) == 2
        ):  # choose to heat up heating water storage
            production_var = self.adding_all_possible_mass_flows(stsv)
            result_hw = self.calculate_new_storage_temperature(
                seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep,
                temperature_sp=temperature_sp_var_hw,
                production=production_var,
                last=last_var_hw,
                c_w=self.specific_heat_capacity_water,
                volume_sp=self.volume_sp_heating_water,
            )

            temperature_sp_c = result_hw[0]
            production_var = 0
            result_ww = self.calculate_new_storage_temperature(
                seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep,
                temperature_sp=temperature_sp_var_ww,
                production=production_var,
                last=last_var_ww,
                c_w=self.specific_heat_capacity_water,
                volume_sp=self.volume_sp_warm_water,
            )

        self.state.temperature_sp_ww = result_ww[0]
        self.state.temperature_sp_hw = result_hw[0]
        stsv.set_output_value(
            self.water_output_temperature_warm_water_channel,
            self.state.temperature_sp_ww,
        )
        stsv.set_output_value(
            self.water_output_temperature_heating_water_channel,
            self.state.temperature_sp_hw,
        )
        stsv.set_output_value(self.water_output_temperature_storage_for_heaters_channel, temperature_sp_c)
        stsv.set_output_value(self.storage_energy_loss_channel, result_ww[1] + result_hw[1])
        stsv.set_output_value(self.real_heat_for_building_channel, last_var_hw)

        # Output Massenstrom von Wasser entspricht dem Input Massenstrom. Nur Temperatur hat sich geändert. Wie ist das zu behandelN?


class HeatStorageController(cp.Component):
    """HeatStorageController class.

    Calculates on base of the maximal Building
    Thermal Demand and the TemperatureHeatingStorage and Building Tempreature
    the real thermal demand for the Heating Storage.
    This Output is called "RealThermalDemandHeatingWater".

    Parameters
    ----------
    sim_params : Simulator
        Simulator object used to carry the simulation using this class

    """

    # Inputs
    # ReferenceMaxHeatBuildingDemand = "ReferenceMaxHeatBuildingDemand"
    TemperatureHeatingStorage = "TemperatureHeatingStorage"
    BuildingTemperature = "BuildingTemperature"
    RealHeatBuildingDemand = "RealHeatBuildingDemand"
    # Outputs
    RealThermalDemandHeatingWater = "RealThermalDemandHeatingWater"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: HeatStorageControllerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.heatstoragecontroller_config = config
        super().__init__(
            name=self.heatstoragecontroller_config.name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.initial_temperature_heating_storage = self.heatstoragecontroller_config.initial_temperature_heating_storage
        self.initial_temperature_building = self.heatstoragecontroller_config.initial_temperature_building

        self.ref_max_thermal_build_demand_in_watt = self.heatstoragecontroller_config.heating_load_of_building_in_watt
        # ===================================================================================================================
        # Inputs
        # self.ref_max_thermal_build_demand: ComponentInput = self.add_input(
        #     self.component_name,
        #     self.ReferenceMaxHeatBuildingDemand,
        #     lt.LoadTypes.HEATING,
        #     lt.Units.WATT,
        #     False,
        # )
        self.heating_storage_temperature_channel: ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureHeatingStorage,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            False,
        )
        self.building_temperature_channel: ComponentInput = self.add_input(
            self.component_name,
            self.BuildingTemperature,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            False,
        )
        self.real_thermal_demand_building_channel = self.add_input(
            self.component_name,
            self.RealHeatBuildingDemand,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            False,
        )
        # Outputs
        self.real_thermal_demand_heating_water_channel: ComponentOutput = self.add_output(
            self.component_name,
            self.RealThermalDemandHeatingWater,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            output_description=f"here a description for {self.RealThermalDemandHeatingWater} will follow.",
        )

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def build(self):
        """Build function."""
        pass

    def write_to_report(self) -> List[str]:
        """Write to report."""
        lines = []
        lines.append("Name: HeatWaterStorage Controller")
        lines.append(f"Initial Temperature Building [°C]: {self.initial_temperature_building}")
        lines.append(f"Initial Temperature Heat Water Storage [°C]: {self.initial_temperature_heating_storage}")
        return self.heatstoragecontroller_config.get_string_dict() + lines

    def i_save_state(self) -> None:
        """Saves the state."""
        pass

    def i_restore_state(self) -> None:
        """Restores the state."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        temperature_sp_var_hw = stsv.get_input_value(self.heating_storage_temperature_channel)  # Start-Temp-Storage
        last_var_hw = stsv.get_input_value(self.real_thermal_demand_building_channel)
        max_mass_flow_heat_storage = self.ref_max_thermal_build_demand_in_watt / (
            4.1851 * 1000 * (self.initial_temperature_heating_storage - self.initial_temperature_building)
        )

        max_last_var_hw = (
            max_mass_flow_heat_storage
            * 4.185
            * 1000
            * (temperature_sp_var_hw - stsv.get_input_value(self.building_temperature_channel))
        )

        if max_last_var_hw < last_var_hw:
            last_var_hw = max_last_var_hw

        stsv.set_output_value(self.real_thermal_demand_heating_water_channel, last_var_hw)



"""Simple Car (LPG connected) and configuration.

Evaluates diesel or electricity consumption based on driven kilometers and processes Car Location for charging stations.
"""

# clean

import datetime as dt
from dataclasses import dataclass

# -*- coding: utf-8 -*-
from typing import List, Any, Tuple, Dict
import numpy as np
import pandas as pd
from dataclasses_json import dataclass_json

from hisim import component as cp
from hisim import loadtypes as lt
from hisim import utils, log
from hisim.component import OpexCostDataClass, CapexCostDataClass
from hisim.components.configuration import EmissionFactorsAndCostsForFuelsConfig
from hisim.simulationparameters import SimulationParameters
from hisim.components.loadprofilegenerator_utsp_connector import UtspLpgConnector
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiHelperClass, KpiTagEnumClass

__authors__ = "Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Johanna Ganglbauer"
__email__ = "johanna.ganglbauer@4wardenergy.at"
__status__ = "development"


@dataclass_json
@dataclass
class GenericCarInformation:

    """Class for collecting important generic car parameters from occupancy."""

    def __init__(self, my_occupancy_instance: UtspLpgConnector):
        """Initialize the class."""

        self.my_occupancy_instance = my_occupancy_instance
        self.build(my_occupancy_instance=my_occupancy_instance)
        self.data_dict_for_car_component = self.prepare_data_dict_for_car_component(
            car_names=self.car_names,
            household_names=self.household_names,
            time_resolutions=self.time_resolutions,
            car_locations=self.car_location_value_list,
            driven_meters=self.driven_meters,
        )

    def build(self, my_occupancy_instance: UtspLpgConnector) -> None:
        """Get important values from occupancy instance."""
        # get names of all available cars
        car_data_dict = my_occupancy_instance.car_data_dict

        # get car names and household names
        (
            self.car_names,
            self.household_names,
            self.time_resolutions,
            self.car_location_value_list,
        ) = self.get_important_parameters_from_occupancy_car_data(car_data_dict=car_data_dict)

        # get driven meters
        self.driven_meters = self.get_meters_driven_from_occupancy_car_data(car_data_dict=car_data_dict)

    def get_important_parameters_from_occupancy_car_data(self, car_data_dict: Dict) -> Tuple[List, List, List, List]:
        """Get car names and household names from occupancy car data."""
        car_location_list = car_data_dict["car_locations"]

        car_names = []
        household_names = []
        time_resolutions = []
        car_location_value_list = []
        for car_location in car_location_list:
            # get car names
            car_name = (
                car_location["LoadTypeName"]
                .split(" - ")[1]
                .translate(str.maketrans({" ": "_", ",": "", "/": "", ".": ""}))
            )
            car_names.append(car_name)

            # get household names
            household_key_dict = car_location["HouseKey"]
            household_name = household_key_dict["HouseholdName"].translate(
                str.maketrans({" ": "_", ",": "", "/": "", ".": ""})
            )
            household_names.append(household_name)

            # get time resolutions
            time_resolution = car_location["TimeResolution"]
            time_resolutions.append(time_resolution)

            # get car location values
            car_location_values = car_location["Values"]
            car_location_value_list.append(car_location_values)

        return car_names, household_names, time_resolutions, car_location_value_list

    def get_meters_driven_from_occupancy_car_data(self, car_data_dict: Dict) -> List:
        """Get meters driven from occupancy car data."""
        driven_distances_list = car_data_dict["driving_distances"]
        driven_meter_list = []
        for driven_distance_data in driven_distances_list:
            driven_meter_values = driven_distance_data["Values"]
            driven_meter_list.append(driven_meter_values)

        return driven_meter_list

    def prepare_data_dict_for_car_component(
        self, car_names: List, household_names: List, time_resolutions: List, car_locations: List, driven_meters: List
    ) -> Dict:
        """Prepare data for car component."""
        data_dict_for_car_component: Dict = {}
        for index, household_name in enumerate(household_names):
            data_dict_for_car_component.update(
                {
                    household_name: {
                        "car_name": car_names[index],
                        "household_name": household_name,
                        "time_resolution": time_resolutions[index],
                        "car_location": car_locations[index],
                        "driven_meters": driven_meters[index],
                    }
                }
            )
        return data_dict_for_car_component


@dataclass_json
@dataclass
class CarConfig(cp.ConfigBase):
    """Definition of configuration of Car."""

    building_name: str
    #: name of the car
    name: str
    #: priority of the component in hierachy: the higher the number the lower the priority
    source_weight: int
    #: type of fuel, either Electricity or Diesel
    fuel: lt.LoadTypes
    #: consumption per kilometer driven, either in kWh/km or l/km
    consumption_per_km: float
    #: CO2 footprint of investment in kg
    co2_footprint: float
    #: cost for investment in Euro
    cost: float
    #: lifetime of car in years
    lifetime: float
    # maintenance cost as share of investment [0..1]
    maintenance_cost_as_percentage_of_investment: float
    #: consumption of the car in kWh or l

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return Car.get_full_classname()

    @classmethod
    def get_default_diesel_config(
        cls,
        name: str = "Car",
        building_name: str = "BUI1",
    ) -> Any:
        """Defines default configuration for diesel vehicle."""
        config = CarConfig(
            building_name=building_name,
            name=name,
            source_weight=1,
            fuel=lt.LoadTypes.DIESEL,
            consumption_per_km=0.06,
            co2_footprint=9139.3,
            cost=32035.0,
            lifetime=18,
            maintenance_cost_as_percentage_of_investment=0.02,
        )
        return config

    @classmethod
    def get_default_ev_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Defines default configuration for electric vehicle."""
        config = CarConfig(
            building_name=building_name,
            name="Car",
            source_weight=1,
            fuel=lt.LoadTypes.ELECTRICITY,
            consumption_per_km=0.15,
            co2_footprint=8899.4,
            cost=44498.0,
            maintenance_cost_as_percentage_of_investment=0.02,
            lifetime=18,
        )
        return config


def most_frequent(input_list: List) -> Any:
    """Returns most frequent value - needed for down sampling Location information from 1 minute resoultion to lower."""
    counter = 0
    num = input_list[0]

    for i in input_list:
        curr_frequency = input_list.count(i)
        if curr_frequency > counter:
            counter = curr_frequency
            num = i
    return num


class Car(cp.Component):
    """Simulates car with constant consumption. Car usage (driven kilometers and state) orginate from LPG."""

    # Outputs
    FuelConsumption = "FuelConsumption"
    ElectricityOutput = "ElectricityOutput"
    CarLocation = "CarLocation"
    DrivenMeters = "DrivenMeters"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: CarConfig,
        data_dict_with_car_information: Dict,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(display_in_webtool=True),
    ) -> None:
        """Initializes Car."""
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.build(config=config, car_information_dict=data_dict_with_car_information)

        if self.config.fuel == lt.LoadTypes.ELECTRICITY:
            self.electricity_output: cp.ComponentOutput = self.add_output(
                object_name=self.component_name,
                field_name=self.ElectricityOutput,
                load_type=lt.LoadTypes.ELECTRICITY,
                unit=lt.Units.WATT,
                postprocessing_flag=[lt.ComponentType.CAR, lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
                output_description="Electricity Consumption of the car while driving. [W]",
            )
            self.car_location_output: cp.ComponentOutput = self.add_output(
                object_name=self.component_name,
                field_name=self.CarLocation,
                load_type=lt.LoadTypes.ANY,
                unit=lt.Units.ANY,
                output_description="Location of the car as integer.",
            )
        elif self.config.fuel == lt.LoadTypes.DIESEL:
            self.fuel_consumption: cp.ComponentOutput = self.add_output(
                object_name=self.component_name,
                field_name=self.FuelConsumption,
                load_type=lt.LoadTypes.DIESEL,
                unit=lt.Units.LITER,
                postprocessing_flag=[
                    lt.InandOutputType.FUEL_CONSUMPTION,
                    lt.LoadTypes.DIESEL,
                    lt.ComponentType.CAR,
                    lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
                ],
                output_description="Diesel Consumption of the car while driving [l].",
            )
        self.driven_meters_output: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.DrivenMeters,
            load_type=lt.LoadTypes.ANY,
            unit=lt.Units.METER,
            output_description="Driven distance in meters.",
        )

    def i_save_state(self) -> None:
        """Saves actual state."""
        pass

    def i_restore_state(self) -> None:
        """Restores previous state."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Checks statements."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Returns consumption and location of car in each timestep."""

        if self.config.fuel == lt.LoadTypes.ELECTRICITY:
            watt_used = (
                self.meters_driven[timestep]
                * self.config.consumption_per_km
                * (3600 / self.my_simulation_parameters.seconds_per_timestep)
            )  # conversion Wh to W
            stsv.set_output_value(self.electricity_output, watt_used)
            stsv.set_output_value(self.car_location_output, self.car_location[timestep])

        # if not already running: check if activation makes sense
        elif self.config.fuel == lt.LoadTypes.DIESEL:
            liters_used = (
                self.meters_driven[timestep] * self.config.consumption_per_km * 1e-3
            )  # conversion meter to kilometer
            stsv.set_output_value(self.fuel_consumption, liters_used)
        stsv.set_output_value(self.driven_meters_output, self.meters_driven[timestep])

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of energy and maintenance costs."""
        co2_per_simulated_period_in_kg = None
        consumption_in_kwh: float
        consumption_in_liter: float
        energy_costs_in_euro = 0.0
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name:
                if (
                        output.field_name == self.FuelConsumption
                        and output.unit == lt.Units.LITER
                        and output.load_type == lt.LoadTypes.DIESEL
                ):
                    consumption_in_liter = round(sum(postprocessing_results.iloc[:, index]), 1)
                    # heating value: https://nachhaltigmobil.schule/leistung-energie-verbrauch/#:~:text=Benzin%20hat%20einen%20Heizwert%20von,9%2C8%20kWh%20pro%20Liter.
                    heating_value_of_diesel_in_kwh_per_liter = 9.8
                    consumption_in_kwh = heating_value_of_diesel_in_kwh_per_liter * consumption_in_liter

                    emissions_and_cost_factors = EmissionFactorsAndCostsForFuelsConfig.get_values_for_year(
                        self.my_simulation_parameters.year
                    )
                    co2_per_unit = emissions_and_cost_factors.diesel_footprint_in_kg_per_l
                    euro_per_unit = emissions_and_cost_factors.diesel_costs_in_euro_per_l

                    energy_costs_in_euro = consumption_in_liter * euro_per_unit
                    co2_per_simulated_period_in_kg = consumption_in_liter * co2_per_unit

                elif (
                    output.field_name == self.ElectricityOutput
                    and output.unit == lt.Units.WATT
                    and output.load_type == lt.LoadTypes.ELECTRICITY
                ):
                    consumption_in_kwh = round(
                        KpiHelperClass.compute_total_energy_from_power_timeseries(
                            power_timeseries_in_watt=postprocessing_results.iloc[:, index],
                            timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                        ),
                        1,
                    )
                    consumption_in_liter = 0
                    # No electricity costs for components except for Electricity Meter, because part of electricity consumption is feed by PV
                    energy_costs_in_euro = 0
                    co2_per_simulated_period_in_kg = 0.0

        if co2_per_simulated_period_in_kg is None:
            raise ValueError("Could not calculate OPEX for Car component.")

        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=energy_costs_in_euro,
            opex_maintenance_cost_in_euro=self.calc_maintenance_cost(),
            co2_footprint_in_kg=co2_per_simulated_period_in_kg,
            consumption_in_kwh=consumption_in_kwh,
            loadtype=self.config.fuel,
            kpi_tag=KpiTagEnumClass.CAR
        )

        return opex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""

        list_of_kpi_entries: List[KpiEntry] = []
        for index, output in enumerate(all_outputs):
            if (
                output.component_name == self.component_name
                and output.field_name == self.ElectricityOutput
                and output.load_type == lt.LoadTypes.ELECTRICITY
            ):
                total_electricity_demand_in_kilowatt_hour = round(
                    KpiHelperClass.compute_total_energy_from_power_timeseries(
                        power_timeseries_in_watt=postprocessing_results.iloc[:, index],
                        timeresolution=self.my_simulation_parameters.seconds_per_timestep,
                    ),
                    1,
                )
                my_kpi_entry = KpiEntry(
                    name="Electricity demand for driving",
                    unit="kWh",
                    value=total_electricity_demand_in_kilowatt_hour,
                    tag=KpiTagEnumClass.CAR,
                    description=self.component_name,
                )
                list_of_kpi_entries.append(my_kpi_entry)
                break
            if (
                output.component_name == self.component_name
                and output.field_name == self.FuelConsumption
                and output.load_type == lt.LoadTypes.DIESEL
                and output.unit == lt.Units.LITER
            ):
                consumption_in_liter = round(sum(postprocessing_results.iloc[:, index]), 1)
                # heating value: https://nachhaltigmobil.schule/leistung-energie-verbrauch/#:~:text=Benzin%20hat%20einen%20Heizwert%20von,9%2C8%20kWh%20pro%20Liter.
                heating_value_of_diesel_in_kwh_per_liter = 9.8
                consumption_in_kwh = round((heating_value_of_diesel_in_kwh_per_liter * consumption_in_liter), 1)

                my_kpi_entry = KpiEntry(
                    name="Diesel demand for driving",
                    unit="liter",
                    value=consumption_in_liter,
                    tag=KpiTagEnumClass.CAR,
                    description=self.component_name,
                )
                list_of_kpi_entries.append(my_kpi_entry)
                my_kpi_entry_2 = KpiEntry(
                    name="Diesel demand for driving",
                    unit="kWh",
                    value=consumption_in_kwh,
                    tag=KpiTagEnumClass.CAR,
                    description=self.component_name,
                )
                list_of_kpi_entries.append(my_kpi_entry_2)
                break

        distance_driven_in_km = round(sum(self.meters_driven) / 1000, 1)
        my_kpi_entry_3 = KpiEntry(
            name="Distance driven",
            unit="km",
            value=distance_driven_in_km,
            tag=KpiTagEnumClass.CAR,
            description=self.component_name,
        )
        list_of_kpi_entries.append(my_kpi_entry_3)

        return list_of_kpi_entries

    @staticmethod
    def get_cost_capex(config: CarConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:
        """Returns investment cost, CO2 emissions and lifetime."""
        seconds_per_year = 365 * 24 * 60 * 60
        capex_per_simulated_period = (config.cost / config.lifetime) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )
        device_co2_footprint_per_simulated_period = (config.co2_footprint / config.lifetime) * (
            simulation_parameters.duration.total_seconds() / seconds_per_year
        )

        capex_cost_data_class = CapexCostDataClass(
            capex_investment_cost_in_euro=config.cost,
            device_co2_footprint_in_kg=config.co2_footprint,
            lifetime_in_years=config.lifetime,
            capex_investment_cost_for_simulated_period_in_euro=capex_per_simulated_period,
            device_co2_footprint_for_simulated_period_in_kg=device_co2_footprint_per_simulated_period,
            kpi_tag=KpiTagEnumClass.CAR,
        )
        return capex_cost_data_class

    def build(self, config: CarConfig, car_information_dict: Dict) -> None:
        """Loads necesary data and saves config to class."""
        self.car_information_dict = car_information_dict
        self.car_location = car_information_dict["car_location"]
        self.meters_driven = car_information_dict["driven_meters"]
        self.time_resolution = car_information_dict["time_resolution"]

        location_translator = {
            "School": 0,
            "Event Location": 0,
            "Shopping": 0,
            None: 0,
            "Home": 1,
            "Workplace": 2,
        }

        # check if caching is possible
        file_exists, cache_filepath = utils.get_cache_file(
            component_key=self.component_name,
            parameter_class=config,
            my_simulation_parameters=self.my_simulation_parameters,
        )
        if file_exists:
            # load from cache
            log.information("Generic car data is taken from cache.")
            dataframe = pd.read_csv(cache_filepath, sep=",", decimal=".", encoding="cp1252")
            self.car_location = dataframe["car_location"].tolist()
            self.meters_driven = dataframe["meters_driven"].tolist()

        else:
            # compare time resolution of LPG to time resolution of hisim
            time_resolution_original = dt.datetime.strptime(self.time_resolution, "%H:%M:%S")
            seconds_per_timestep_original = (
                time_resolution_original.hour * 3600
                + time_resolution_original.minute * 60
                + time_resolution_original.second
            )
            minutes_per_timestep = int(
                self.my_simulation_parameters.seconds_per_timestep / seconds_per_timestep_original
            )

            simulation_time_span = self.my_simulation_parameters.end_date - self.my_simulation_parameters.start_date
            # minutes_per_timestep = int(self.my_simulation_parameters.seconds_per_timestep / 60)
            steps_desired = self.my_simulation_parameters.timesteps
            steps_desired_in_minutes = steps_desired * minutes_per_timestep

            # extract values for location and distance of car,
            # include time information and
            # translate car location to integers (according to location_translator)
            initial_data = pd.DataFrame(
                {
                    "Time": pd.date_range(
                        start=dt.datetime(year=self.my_simulation_parameters.year, month=1, day=1),
                        end=dt.datetime(year=self.my_simulation_parameters.year, month=1, day=1)
                        + dt.timedelta(days=simulation_time_span.days)
                        - dt.timedelta(seconds=60),
                        freq="T",
                    ),
                    "meters_driven": self.meters_driven[:steps_desired_in_minutes],
                    "car_location": [location_translator[elem] for elem in self.car_location][
                        :steps_desired_in_minutes
                    ],
                }
            )
            initial_data = utils.convert_lpg_data_to_utc(data=initial_data, year=self.my_simulation_parameters.year)
            self.meters_driven = pd.to_numeric(initial_data["meters_driven"]).tolist()
            self.car_location = pd.to_numeric(initial_data["car_location"]).tolist()

            # sum / extract most common value from data to match hisim time resolution
            if minutes_per_timestep > 1:
                self.meters_driven = self.resample_meters_driven(
                    meters_driven=self.meters_driven,
                    seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep,
                )
                self.car_location = [
                    most_frequent(
                        input_list=self.car_location[i * minutes_per_timestep : (i + 1) * minutes_per_timestep]
                    )
                    for i in range(steps_desired)
                ]
            else:
                self.meters_driven = self.meters_driven
                self.car_location = self.car_location

            # save data in cache
            database = pd.DataFrame({"car_location": self.car_location, "meters_driven": self.meters_driven})
            database.to_csv(cache_filepath)
            del database

    def resample_meters_driven(self, meters_driven: List, seconds_per_timestep: int) -> Any:
        """Resample meters driven according to simulation time resolution."""
        # Convert seconds per timestep to minutes per timestep
        minutes_per_timestep = seconds_per_timestep // 60

        # Check the length of the input list
        total_minutes = len(meters_driven)

        # Calculate the number of complete timesteps
        num_timesteps = total_minutes // minutes_per_timestep

        # Trim the list to be a multiple of minutes_per_timestep
        trimmed_meters_driven = meters_driven[: num_timesteps * minutes_per_timestep]

        # Reshape and sum the data
        reshaped_meters = np.reshape(trimmed_meters_driven, (num_timesteps, minutes_per_timestep))
        resampled_meters = np.sum(reshaped_meters, axis=1)
        return resampled_meters

    def write_to_report(self) -> List[str]:
        """Writes Car values to report."""
        return self.config.get_string_dict()



"""Model Predictive Controller."""

# clean

import datetime
from typing import List, Optional, Any

# from typing import Any
from dataclasses import dataclass

# from statistics import mean
import numpy as np

# from numpy.linalg import inv
from dataclasses_json import dataclass_json

# from scipy.ndimage import interpolation
import casadi as ca

# Owned
from hisim import utils
from hisim import component as cp
from hisim.component import ConfigBase
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters
from hisim.components.weather import Weather

# from hisim.components.generic_battery import GenericBattery
# from hisim.components.loadprofilegenerator_connector import Occupancy
# from hisim.components.generic_price_signal import PriceSignal
# from hisim.components.air_conditioner import AirConditioner
# from hisim.components.generic_pv_system import PVSystem
from hisim import log
from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum

__authors__ = "Marwa Alfouly"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Dr. Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class MpcControllerConfig(ConfigBase):
    """Configuration of the MPC Controller."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return MpcController.get_full_classname()

    building_name: str
    # parameter_string: str
    # my_simulation_parameters: SimulationParameters
    name: str
    mpc_scheme: str
    min_comfort_temp: float
    max_comfort_temp: float
    optimizer_sampling_rate: int
    initial_temeperature: float
    flexibility_element: str
    initial_state_of_charge: float
    # my_simulation_repository: Optional[ cp.SimRepository ]
    # getting forecasted disturbance (weather)
    temp_forecast: List[float]
    phi_m_forecast: List[float]
    phi_st_forecast: list
    phi_ia_forecast: list
    # getting pv forecast
    pv_forecast_yearly: list
    # getting battery specifications
    maximum_storage_capacity: float
    minimum_storage_capacity: float
    maximum_charging_power: float
    maximum_discharging_power: float
    battery_efficiency: float
    inverter_efficiency: float
    # forecasts
    temperature_forecast_24h_1min: list
    phi_m_forecast_24h_1min: list
    phi_ia_forecast_24h_1min: list
    phi_st_forecast_24h_1min: list
    pv_forecast_24h_1min: list
    price_purchase_forecast_24h_1min: list
    price_injection_forecast_24h_1min: list
    optimal_cost: list
    revenues: list
    air_conditioning_electricity: list
    cost_optimal_temperature_set_point: list
    pv2load: list
    electricity_from_grid: list
    electricity_to_grid: list
    battery_to_load: list
    pv_to_battery_timestep: list
    battery_power_flow_timestep: list
    battery_control_state: list
    batt_soc_actual_timestep: list
    batt_soc_normalized_timestep: list
    # thermal building model
    h_tr_w: float
    h_tr_ms: float
    h_tr_em: float
    h_ve_adj: float
    h_tr_is: float
    c_m: float
    cop_coef: List
    eer_coef: List
    predictive: bool
    prediction_horizon: Optional[int]

    @classmethod
    def get_default_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default MPC controller."""
        return MpcControllerConfig(
            building_name=building_name,
            name="MpcController",
            mpc_scheme="optimization_once_aday_only",
            min_comfort_temp=21.0,
            max_comfort_temp=23.0,
            optimizer_sampling_rate=15,
            initial_temeperature=22.0,
            flexibility_element="basic_buidling_configuration",
            initial_state_of_charge=10 / 15,
            # my_simulation_repository = [],
            # getting forecasted disturbance (weather)
            temp_forecast=[],
            phi_m_forecast=[],
            phi_st_forecast=[],
            phi_ia_forecast=[],
            # getting pv forecast
            pv_forecast_yearly=[],
            # getting battery specifications
            maximum_storage_capacity=0.0,
            minimum_storage_capacity=0.0,
            maximum_charging_power=0.0,
            maximum_discharging_power=0.0,
            battery_efficiency=0.0,
            inverter_efficiency=0.0,
            # forecasts
            temperature_forecast_24h_1min=[],
            phi_m_forecast_24h_1min=[],
            phi_ia_forecast_24h_1min=[],
            phi_st_forecast_24h_1min=[],
            pv_forecast_24h_1min=[],
            price_purchase_forecast_24h_1min=[],
            price_injection_forecast_24h_1min=[],
            optimal_cost=[],
            revenues=[],
            air_conditioning_electricity=[],
            cost_optimal_temperature_set_point=[],
            pv2load=[],
            electricity_from_grid=[],
            electricity_to_grid=[],
            battery_to_load=[],
            pv_to_battery_timestep=[],
            battery_power_flow_timestep=[],
            battery_control_state=[],
            batt_soc_actual_timestep=[],
            batt_soc_normalized_timestep=[],
            h_tr_w=0.0,
            h_tr_ms=0.0,
            h_tr_em=0.0,
            h_ve_adj=0.0,
            h_tr_is=0.0,
            c_m=0.0,
            cop_coef=[0] * 2,
            eer_coef=[0] * 2,
            predictive=True,
            prediction_horizon=0,
        )


class MPCcontrollerState:
    """Controller state."""

    def __init__(self, t_m: float, soc: float, cost_optimal_thermal_power: list):
        """Constructs all the neccessary attributes for the MPCcontrollerState object."""
        self.t_m: float = t_m
        self.soc: float = soc
        self.cost_optimal_thermal_power: list = cost_optimal_thermal_power

    def clone(self):
        """Copies the Controller State."""
        return MPCcontrollerState(self.t_m, self.soc, self.cost_optimal_thermal_power)


class MpcController(cp.Component):
    """MPC Controller class."""

    # Inputs
    # weather
    TemperatureOutside = "TemperatureOutside"
    # building
    TemperatureMean = "Residence Temperature"

    # Outputs
    TemperatureMeanStateSpace = "TemperatureMeanStateSpace"
    TemperatureSurfaceStateSpace = "TemperatureSurfaceStateSpace"
    TemperatureAirStateSpace = "TemperatureAirStateSpace"
    ThermalEnergyDelivered = "ThermalEnergyDelivered"
    ElectricityOutput = "ElectricityOutput"
    OperatingMode = "OperatingMode"
    PV2load = "PV2load"
    GridImport = "GridImport"
    GridExport = "GridExport"
    Battery2Load = "Battery2Load"
    PV2Battery = "PV2Battery"
    BatteryChargingDischargingPower = "BatteryChargingDischargingPower"
    BatteryControlState = "BatteryControlState"
    BatteryEnergyContent = "BatteryEnergyContent"
    BatterySoC = "BatterySoC"
    ElectricityCost = "ElectricityCost"
    GenerationRevenue = "GenerationRevenue"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: MpcControllerConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Constructs all the neccessary attributes."""

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.my_simulation_parameters = my_simulation_parameters

        self.mpcconfig = config

        self.h_tr_w = self.mpcconfig.h_tr_w
        self.h_tr_ms = self.mpcconfig.h_tr_ms
        self.h_tr_em = self.mpcconfig.h_tr_em
        self.h_ve_adj = self.mpcconfig.h_ve_adj
        self.h_tr_is = self.mpcconfig.h_tr_is
        self.c_m = self.mpcconfig.c_m
        self.cop_coef = self.mpcconfig.cop_coef
        self.eer_coef = self.mpcconfig.eer_coef

        self.build()

        self.statespace()

        self.t_m_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureMean,
            LoadTypes.TEMPERATURE,
            Units.CELSIUS,
            True,
        )

        self.operating_mode_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.OperatingMode,
            LoadTypes.ANY,
            Units.ANY,
            output_description=f"here a description for {self.OperatingMode} will follow.",
        )

        self.p_th_mpc_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ThermalEnergyDelivered,
            LoadTypes.HEATING,
            Units.WATT,
            output_description=f"here a description for {self.ThermalEnergyDelivered} will follow.",
        )
        self.p_elec_mpc_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ElectricityOutput,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for {self.ElectricityOutput} will follow.",
        )

        self.pv_consumption_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.PV2load,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for {self.PV2load} will follow.",
        )

        self.grid_import_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.GridImport,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for {self.GridImport} will follow.",
        )
        self.grid_export_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.GridExport,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for {self.GridExport} will follow.",
        )

        self.battery_to_load_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.Battery2Load,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for {self.Battery2Load} will follow.",
        )

        self.pv_to_battery_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.PV2Battery,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for {self.PV2Battery} will follow.",
        )

        self.batt_soc_actual_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.BatteryEnergyContent,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for {self.BatteryEnergyContent} will follow.",
        )
        self.battery_control_state_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.BatteryControlState,
            LoadTypes.ANY,
            Units.ANY,
            output_description=f"here a description for {self.BatteryControlState} will follow.",
        )
        self.battery_charging_discharging_power_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.BatteryChargingDischargingPower,
            LoadTypes.ELECTRICITY,
            Units.WATT,
            output_description=f"here a description for {self.BatteryChargingDischargingPower} will follow.",
        )

        self.batt_soc_normalized_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.BatterySoC,
            LoadTypes.ANY,
            Units.ANY,
            output_description=f"here a description for {self.BatterySoC} will follow.",
        )

        self.costs_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.ElectricityCost,
            LoadTypes.PRICE,
            Units.EUR_PER_KWH,
            output_description=f"here a description for {self.ElectricityCost} will follow.",
        )

        self.revenues_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.GenerationRevenue,
            LoadTypes.PRICE,
            Units.EUR_PER_KWH,
            output_description=f"here a description for {self.GenerationRevenue} will follow.",
        )

        self.add_default_connections(self.get_weather_default_connections())

        if self.mpcconfig.prediction_horizon is not None:
            self.prediction_horizon = int(
                self.mpcconfig.prediction_horizon / self.my_simulation_parameters.seconds_per_timestep
            )
        self.state: MPCcontrollerState = MPCcontrollerState(
            t_m=self.mpcconfig.initial_temeperature,
            soc=self.mpcconfig.initial_state_of_charge,
            cost_optimal_thermal_power=self.prediction_horizon * [0],
        )
        self.previous_state = self.state.clone()

        self.mpc_scheme = self.mpcconfig.mpc_scheme
        self.min_comfort_temp = self.mpcconfig.min_comfort_temp
        self.max_comfort_temp = self.mpcconfig.max_comfort_temp
        self.sampling_rate = self.mpcconfig.optimizer_sampling_rate
        self.flexibility_element = self.mpcconfig.flexibility_element

        self.temp_forecast = self.mpcconfig.temp_forecast
        self.phi_m_forecast = self.mpcconfig.phi_m_forecast
        self.phi_st_forecast = self.mpcconfig.phi_st_forecast
        self.phi_ia_forecast = self.mpcconfig.phi_ia_forecast
        self.pv_forecast_yearly = self.mpcconfig.pv_forecast_yearly
        self.maximum_storage_capacity = self.mpcconfig.maximum_storage_capacity
        self.minimum_storage_capacity = self.mpcconfig.minimum_storage_capacity
        self.maximum_charging_power = self.mpcconfig.maximum_charging_power
        self.maximum_discharging_power = self.mpcconfig.maximum_discharging_power
        self.battery_efficiency = self.mpcconfig.battery_efficiency
        self.inverter_efficiency = self.mpcconfig.inverter_efficiency

        self.temperature_forecast_24h_1min = self.mpcconfig.temperature_forecast_24h_1min
        self.phi_m_forecast_24h_1min = self.mpcconfig.phi_m_forecast_24h_1min
        self.phi_ia_forecast_24h_1min = self.mpcconfig.phi_ia_forecast_24h_1min
        self.phi_st_forecast_24h_1min = self.mpcconfig.phi_st_forecast_24h_1min
        self.pv_forecast_24h_1min = self.mpcconfig.pv_forecast_24h_1min
        self.price_purchase_forecast_24h_1min = self.mpcconfig.price_purchase_forecast_24h_1min
        self.price_injection_forecast_24h_1min = self.mpcconfig.price_injection_forecast_24h_1min
        self.optimal_cost = self.mpcconfig.optimal_cost
        self.revenues = self.mpcconfig.revenues
        self.air_conditioning_electricity = self.mpcconfig.air_conditioning_electricity
        self.cost_optimal_temperature_set_point = self.mpcconfig.cost_optimal_temperature_set_point
        self.pv2load = self.mpcconfig.pv2load
        self.electricity_from_grid = self.mpcconfig.electricity_from_grid
        self.electricity_to_grid = self.mpcconfig.electricity_to_grid
        self.battery_to_load = self.mpcconfig.battery_to_load
        self.pv_to_battery_timestep = self.mpcconfig.pv_to_battery_timestep
        self.battery_power_flow_timestep = self.mpcconfig.battery_power_flow_timestep
        self.battery_control_state = self.mpcconfig.battery_control_state
        self.batt_soc_actual_timestep = self.mpcconfig.batt_soc_actual_timestep
        self.batt_soc_normalized_timestep = self.mpcconfig.batt_soc_normalized_timestep

    def get_weather_default_connections(self):
        """Get default connections from the weather component."""

        connections = []
        weather_classname = Weather.get_classname()
        connections.append(
            cp.ComponentConnection(
                MpcController.TemperatureOutside,
                weather_classname,
                Weather.TemperatureOutside,
            )
        )
        return connections

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        if self.mpcconfig.predictive:
            """Get forecasted disturbance (weather)"""
            self.temp_forecast = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.WEATHERTEMPERATUREOUTSIDEYEARLYFORECAST
            )[: self.my_simulation_parameters.timesteps]
            self.phi_m_forecast = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.HEATFLUXTHERMALMASSNODEFORECAST
            )
            self.phi_st_forecast = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.HEATFLUXSURFACENODEFORECAST
            )
            self.phi_ia_forecast = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.HEATFLUXINDOORAIRNODEFORECAST
            )

            """"getting pv forecast"""
            self.pv_forecast_yearly = SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.PVFORECASTYEARLY)

            """ getting battery specifications """
            self.maximum_storage_capacity = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.MAXIMUMBATTERYCAPACITY
            )
            self.minimum_storage_capacity = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.MINIMUMBATTERYCAPACITY
            )
            self.maximum_charging_power = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.MAXIMALCHARGINGPOWER
            )
            self.maximum_discharging_power = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.MAXIMALDISCHARGINGPOWER
            )
            self.battery_efficiency = SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.BATTERYEFFICIENCY)
            self.inverter_efficiency = SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.INVERTEREFFICIENCY)
            log.information(f"self.inverter_efficiency {format(self.inverter_efficiency)}")

    def build(self):
        """Build function: The function sets important constants and parameters for the calculations."""
        if self.mpcconfig.predictive:
            """getting building physical properties for state space model"""
            self.h_tr_w = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.THERMALTRANSMISSIONCOEFFICIENTGLAZING
            )
            self.h_tr_ms = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.THERMALTRANSMISSIONCOEFFICIENTOPAQUEMS
            )
            self.h_tr_em = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.THERMALTRANSMISSIONCOEFFICIENTOPAQUEEM
            )
            self.h_ve_adj = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.THERMALTRANSMISSIONCOEFFICIENTVENTILLATION
            )
            self.h_tr_is = SingletonSimRepository().get_entry(
                key=SingletonDictKeyEnum.THERMALTRANSMISSIONSURFACEINDOORAIR
            )
            self.c_m = SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.THERMALCAPACITYENVELOPE)
            """"
            self.h_tr_w = my_simulation_repository.get_entry(
                Building.Thermal_transmission_coefficient_glazing
            )
            self.h_tr_ms = my_simulation_repository.get_entry(
                Building.Thermal_transmission_coefficient_opaque_ms
            )
            self.h_tr_em = my_simulation_repository.get_entry(
                Building.Thermal_transmission_coefficient_opaque_em
            )
            self.h_ve_adj = my_simulation_repository.get_entry(
                Building.Thermal_transmission_coefficient_ventillation
            )
            self.h_tr_is = my_simulation_repository.get_entry(
                Building.Thermal_transmission_Surface_IndoorAir
            )
            self.c_m = my_simulation_repository.get_entry(
                Building.Thermal_capacity_envelope
            )
            """

            """ getting cop_coef and eer_coef from the air conditioner omponenent to be used in the cost optimization"""
            self.cop_coef = SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.COPCOEFFICIENTHEATING)
            self.eer_coef = SingletonSimRepository().get_entry(key=SingletonDictKeyEnum.EERCOEFFICIENTCOOLING)

    def statespace(self):
        """State Space Model of the 5R1C network, Used as a prediction model to the building behavior in the MPC."""

        seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep

        state_space_matrix_coefficient_x = ((self.h_tr_w + self.h_tr_ms) * (self.h_ve_adj + self.h_tr_is)) + (
            self.h_ve_adj * self.h_tr_is
        )

        # Entries for system matrix
        a11 = (
            ((self.h_tr_ms**2) * (self.h_tr_is + self.h_ve_adj) / state_space_matrix_coefficient_x)
            - self.h_tr_ms
            - self.h_tr_em
        ) / (
            self.c_m / seconds_per_timestep
        )  # ((self.c_m_ref * self.A_f) * 3600)

        # Entries for input matrix
        b11 = (self.h_tr_ms * self.h_tr_is) / ((self.c_m / seconds_per_timestep) * state_space_matrix_coefficient_x)

        b_d11 = (
            (self.h_tr_ms * self.h_tr_w * (self.h_tr_is + self.h_ve_adj) / state_space_matrix_coefficient_x)
            + self.h_tr_em
        ) / (self.c_m / seconds_per_timestep)
        b_d12 = (self.h_tr_ms * self.h_tr_is * self.h_ve_adj) / (
            (self.c_m / seconds_per_timestep) * state_space_matrix_coefficient_x
        )
        b_d13 = (self.h_tr_ms * self.h_tr_is) / ((self.c_m / seconds_per_timestep) * state_space_matrix_coefficient_x)
        b_d14 = (self.h_tr_ms * (self.h_tr_is + self.h_ve_adj)) / (
            (self.c_m / seconds_per_timestep) * state_space_matrix_coefficient_x
        )
        b_d15 = 1 / (self.c_m / seconds_per_timestep)

        # Entries for output matrix
        c11 = (self.h_tr_ms * self.h_tr_is) / state_space_matrix_coefficient_x
        c21 = (self.h_tr_ms * (self.h_tr_is + self.h_ve_adj)) / state_space_matrix_coefficient_x

        # Entries for feedthrough matrix
        d11 = (self.h_tr_ms + self.h_tr_w + self.h_tr_is) / state_space_matrix_coefficient_x
        d21 = self.h_tr_is / state_space_matrix_coefficient_x

        d_d11 = (self.h_tr_w * self.h_tr_is) / state_space_matrix_coefficient_x
        d_d12 = (self.h_tr_ms + self.h_tr_is + self.h_tr_w) * self.h_ve_adj / state_space_matrix_coefficient_x
        d_d13 = (self.h_tr_ms + self.h_tr_is + self.h_tr_w) / state_space_matrix_coefficient_x
        d_d14 = self.h_tr_is / state_space_matrix_coefficient_x
        d_d15 = 0
        d_d21 = (self.h_tr_w * (self.h_tr_is + self.h_ve_adj)) / state_space_matrix_coefficient_x
        d_d22 = (self.h_tr_is * self.h_ve_adj) / state_space_matrix_coefficient_x
        d_d23 = self.h_tr_is / state_space_matrix_coefficient_x
        d_d24 = (self.h_tr_is + self.h_ve_adj) / state_space_matrix_coefficient_x
        d_d25 = 0

        # Build arrays for state space representation
        state_space_system_matrix_a = np.array([[a11]])  # system matrix
        state_space_input_matrix_b_d = np.array([[b_d11, b_d12, b_d13, b_d14, b_d15]])  # input matrix discrete
        state_space_input_matrix_b = np.array([[b11, b_d11, b_d12, b_d13, b_d14, b_d15]])  # input matrix
        state_space_output_matrix_c = np.array([[c11], [c21]])  # output matrix
        state_space_feedthrough_matrix_d_d = np.array(
            [[d_d11, d_d12, d_d13, d_d14, d_d15], [d_d21, d_d22, d_d23, d_d24, d_d25]]
        )  # feedthrough matrix discrete
        state_space_feedthrough_matrix_d = np.array(
            [
                [d11, d_d11, d_d12, d_d13, d_d14, d_d15],
                [d21, d_d21, d_d22, d_d23, d_d24, d_d25],
            ]
        )  # feedthrough matrix

        self.state_space_system_matrix_a = state_space_system_matrix_a * 0.5
        self.state_space_system_matrix_b = state_space_input_matrix_b * 0.5

        return (
            state_space_system_matrix_a,
            state_space_input_matrix_b,
            state_space_input_matrix_b_d,
            state_space_output_matrix_c,
            state_space_feedthrough_matrix_d,
            state_space_feedthrough_matrix_d_d,
            a11,
            b11,
            b_d11,
            b_d12,
            b_d13,
            b_d14,
            b_d15,
        )

    def i_save_state(self):
        """Saves the current state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self):
        """Restores previous state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def write_to_report(self):
        """Writes a report."""
        lines = []
        lines.append(f"Name: {format('Model Predictive Controller')}")
        lines.append("tbd")
        return lines

    def get_forecast_24h(self, start_horizon, sampling_rate):
        """Get yearly weather forecast."""
        # slicing yearly forecast to extract data points for the prediction horizon (24 hours)
        # number of data points extracted equals= self.prediction_horizon = 3600 * 24 h / seconds_per_timestep
        self.temperature_forecast_24h_1min = self.temp_forecast[start_horizon : start_horizon + self.prediction_horizon]
        self.phi_m_forecast_24h_1min = self.phi_m_forecast[start_horizon : start_horizon + self.prediction_horizon]
        self.phi_ia_forecast_24h_1min = self.phi_ia_forecast[start_horizon : start_horizon + self.prediction_horizon]
        self.phi_st_forecast_24h_1min = self.phi_st_forecast[start_horizon : start_horizon + self.prediction_horizon]
        self.pv_forecast_24h_1min = self.pv_forecast_yearly[start_horizon : start_horizon + self.prediction_horizon]

        self.price_purchase_forecast_24h_1min = SingletonSimRepository().get_entry(
            key=SingletonDictKeyEnum.PRICEPURCHASEFORECAST24H
        )
        self.price_injection_forecast_24h_1min = SingletonSimRepository().get_entry(
            key=SingletonDictKeyEnum.PRICEINJECTIONFORECAST24H
        )

        # sampling of the data: Useful if you run hisim at 60 sec per time step and you want fast optimization:
        # Recommended to use 15 min or 20 min

        temperature_forecast_24h = self.temperature_forecast_24h_1min[0::sampling_rate]
        phi_m_forecast_24h = self.phi_m_forecast_24h_1min[0::sampling_rate]
        phi_ia_forecast_24h = self.phi_ia_forecast_24h_1min[0::sampling_rate]
        phi_st_forecast_24h = self.phi_st_forecast_24h_1min[0::sampling_rate]
        price_purchase_forecast_24h = self.price_purchase_forecast_24h_1min[0::sampling_rate]
        price_injection_forecast_24h = self.price_injection_forecast_24h_1min[0::sampling_rate]
        pv_forecast_24h = self.pv_forecast_24h_1min[0::sampling_rate]

        # only take positive values from PV, otherwise mpc will have an ill-posed problem which will raise errors
        pv_forecast_24h = [max(0, value) for value in pv_forecast_24h]
        return (
            temperature_forecast_24h,
            phi_ia_forecast_24h,
            phi_st_forecast_24h,
            phi_m_forecast_24h,
            price_purchase_forecast_24h,
            price_injection_forecast_24h,
            pv_forecast_24h,
        )

    @utils.measure_execution_time
    def optimize(  # noqa: C901
        self,
        temperature_forecast_24h,
        phi_ia_forecast_24h,
        phi_st_forecast_24h,
        phi_m_forecast_24h,
        price_purchase_forecast_24h,
        price_injection_forecast_24h,
        pv_forecast_24h,
        scaled_horizon,
    ):
        """MPC implementation."""
        sampling_rate = int(self.prediction_horizon / scaled_horizon)
        # scaled_horizon = scaled_horizon  # scaled prediction horizon

        # Discretization of the state space model:
        identity_matrix = np.identity(self.state_space_system_matrix_a.shape[0])  # this is an identity matrix
        matrix_a_d = np.array(np.exp(self.state_space_system_matrix_a * sampling_rate))
        matrix_b_d = (
            np.linalg.inv(self.state_space_system_matrix_a)
            * (matrix_a_d - identity_matrix)
            * self.state_space_system_matrix_b
        )

        # numerical values of the disturbances
        disturbance_values = ca.horzcat(
            temperature_forecast_24h,
            temperature_forecast_24h,
            phi_ia_forecast_24h,
            phi_st_forecast_24h,
            phi_m_forecast_24h,
        ).T

        # Numerical values of cop and eer sampled (casadi fromat)

        cop_timestep = []
        eer_timestep = []
        for k in range(int(self.prediction_horizon)):
            cop_timestep.append(self.cop_coef[0] * self.temperature_forecast_24h_1min[k] + self.cop_coef[1])  # cop
            eer_timestep.append(self.eer_coef[0] * self.temperature_forecast_24h_1min[k] + self.eer_coef[1])  # eer

        cop_sampled = cop_timestep[0::sampling_rate]
        eer_sampled = eer_timestep[0::sampling_rate]

        cop_sampled_array = np.reshape(np.array(cop_sampled), (1, len(cop_sampled)))
        eer_sampled_array = np.reshape(np.array(eer_sampled), (1, len(eer_sampled)))
        cop_sampled_array = np.reshape(np.array(cop_sampled), (1, len(cop_sampled)))
        eer_sampled_array = np.reshape(np.array(eer_sampled), (1, len(eer_sampled)))

        # Numerical values of pv forecast (casadi fromat)
        pv_forecast_24h = np.reshape(np.array(pv_forecast_24h), (1, len(pv_forecast_24h)))

        p_el = np.reshape(np.array(price_purchase_forecast_24h), (1, len(price_purchase_forecast_24h)))

        feed_in_tariff = np.reshape(
            np.array(price_injection_forecast_24h),
            (1, len(price_injection_forecast_24h)),
        )

        # symbolic defenition of system variables:

        # 1. manipulated variable
        phi_hc = ca.MX.sym("phi_hc")

        # 2. state: controlled variable (thermal power delivered)
        t_m = ca.MX.sym("t_m")

        # 3. Disturbances (outside temperature - solar gains - internal gains - dynamic price signal)
        t_out = ca.MX.sym("t_out")
        t_sup = ca.MX.sym("t_sup")
        phi_ia = ca.MX.sym("phi_ia")
        phi_st = ca.MX.sym("phi_st")
        phi_m = ca.MX.sym("phi_m")

        disturbances = ca.vertcat(t_out, t_sup, phi_ia, phi_st, phi_m)
        n_disturbances = disturbances.numel()

        t_m_discrete = (
            matrix_a_d * t_m
            + matrix_b_d[0, 0] * phi_hc
            + matrix_b_d[0, 1] * t_out
            + matrix_b_d[0, 2] * t_sup
            + matrix_b_d[0, 3] * phi_ia
            + matrix_b_d[0, 4] * phi_st
            + matrix_b_d[0, 5] * phi_m
        )

        supporting_points_multiple_shooting_continuity_condition = ca.Function(
            "F",
            [t_m, phi_hc, t_out, t_sup, phi_ia, phi_st, phi_m],
            [t_m_discrete],
            ["t_m", "phi_hc", "t_out", "t_sup", "phi_ia", "phi_st", "phi_m"],
            ["t_m_next"],
        )

        # multiple shooting approach: more than one decision variable (state variables, thermal power, grid import, ...)

        opti = ca.Opti()

        optvar_temperature = opti.variable(1, scaled_horizon + 1)  # state variable: controlled temperature
        optvar_power_thermal_delivered = opti.variable(
            1, scaled_horizon
        )  # manipulated variable: thermal power delivered
        optvar_disturbances = opti.variable(n_disturbances, scaled_horizon)
        # disturbances:
        # 1. ambient temperature
        # 2. supply temperature = ambient temperature
        # 3. heat flux to the node Ti (indoor air)
        # 4. heat flux to the node s (internal surfaces)
        # 5. heat flux to thermal mass node
        optvar_power_bought_from_grid = opti.variable(1, scaled_horizon)

        if self.flexibility_element in {"PV_only", "PV_and_Battery"}:
            optvar_power_pv = opti.variable(1, scaled_horizon)
            optvar_power_sold_to_grid = opti.variable(1, scaled_horizon)
            optvar_power_pv_generation_forecasted = opti.variable(1, scaled_horizon)

        if self.flexibility_element == "PV_and_Battery":
            optvar_battery_soc = opti.variable(1, scaled_horizon + 1)
            optvar_battery_power_charging = opti.variable(1, scaled_horizon)
            optvar_battery_power_discharging = opti.variable(1, scaled_horizon)
            optvar_battery_power_flow = opti.variable(1, scaled_horizon)
            # flow=opti.variable(1,N)

        x_init = opti.parameter(1, 1)
        # u_init=opti.parameter(1,1)
        disturbance_forecast = opti.parameter(n_disturbances, scaled_horizon)
        cop_values = opti.parameter(
            1, scaled_horizon
        )  # coefiiecient of performance: heating air conditioner efficiency
        eer_values = opti.parameter(1, scaled_horizon)  # energy efficiency ratio: cooling air conditioner efficiency

        if self.flexibility_element in {"PV_only", "PV_and_Battery"}:
            pv_production = opti.parameter(1, scaled_horizon)

        if self.flexibility_element == "PV_and_Battery":
            soc_init = opti.parameter(1, 1)

        # Cost Function

        if self.flexibility_element == "basic_buidling_configuration":
            opti.minimize(sum(ca.horzsplit(p_el * optvar_power_bought_from_grid, 1)))

        if self.flexibility_element == "PV_only":
            # a weighting factor of 0.5 is added to the revenue to priortize using the pv production instead of selling to the grid
            opti.minimize(
                sum(
                    ca.horzsplit(
                        (p_el * optvar_power_bought_from_grid - 0.5 * feed_in_tariff * optvar_power_sold_to_grid)
                    )
                )
            )

        if self.flexibility_element == "PV_and_Battery":
            opti.minimize(
                sum(
                    ca.horzsplit(
                        (p_el * optvar_power_bought_from_grid - 0.5 * feed_in_tariff * optvar_power_sold_to_grid)
                    )
                )
            )

        # Constraints
        for k in range(scaled_horizon):
            opti.subject_to(
                optvar_temperature[:, k + 1]
                == supporting_points_multiple_shooting_continuity_condition(
                    optvar_temperature[:, k],
                    optvar_power_thermal_delivered[:, k],
                    disturbance_forecast[0, k],
                    disturbance_forecast[1, k],
                    disturbance_forecast[2, k],
                    disturbance_forecast[3, k],
                    disturbance_forecast[4, k],
                )
            )
            if self.flexibility_element == "PV_and_Battery":
                opti.subject_to(
                    optvar_battery_soc[:, k + 1]
                    == optvar_battery_soc[:, k]
                    + (optvar_battery_power_flow[:, k])
                    * (self.my_simulation_parameters.seconds_per_timestep * sampling_rate / 3600)
                )

        opti.subject_to(opti.bounded(self.min_comfort_temp, optvar_temperature, self.max_comfort_temp))

        """ a Terminal Constraint is added if Hisim resolution is different than the optimizer resolution:
            e.g, if you run hisim at 60 sec per time step and you would like to reduce the optimization is done by sampling each 20 or 15 min
            This ensures that inital guess at the following optimization is within the constraint"""

        if sampling_rate != 1:
            opti.subject_to(optvar_temperature[-1] > self.min_comfort_temp + 0.3)

        opti.subject_to(opti.bounded(0, ca.fabs(optvar_power_thermal_delivered), 16000))

        if self.flexibility_element == "basic_buidling_configuration":
            opti.subject_to(
                optvar_power_bought_from_grid
                == ca.if_else(
                    optvar_power_thermal_delivered > 0,
                    ca.fabs(optvar_power_thermal_delivered) / cop_values,
                    ca.fabs(optvar_power_thermal_delivered) / eer_values,
                )
            )

        if self.flexibility_element == "PV_only":
            """Energy Balance constraint for Grid , PV  interaction"""
            opti.subject_to(
                optvar_power_bought_from_grid
                == ca.if_else(
                    optvar_power_thermal_delivered > 0,
                    ca.fabs(optvar_power_thermal_delivered) / cop_values - (pv_production - optvar_power_sold_to_grid),
                    ca.fabs(optvar_power_thermal_delivered) / eer_values - (pv_production - optvar_power_sold_to_grid),
                )
            )
            opti.subject_to(optvar_power_pv == pv_production - optvar_power_sold_to_grid)

        if self.flexibility_element == "PV_and_Battery":
            """Battery charging and discharging bounds / making sure that charging and discharging doesn't occur at the same time"""
            opti.subject_to(
                opti.bounded(
                    self.minimum_storage_capacity,
                    optvar_battery_soc,
                    self.maximum_storage_capacity,
                )
            )
            opti.subject_to(
                opti.bounded(
                    -self.maximum_discharging_power,
                    optvar_battery_power_flow,
                    self.maximum_charging_power,
                )
            )
            opti.subject_to(
                optvar_battery_power_charging
                == ca.if_else(
                    optvar_battery_power_flow > 0,
                    optvar_battery_power_flow * self.battery_efficiency,
                    0,
                )
            )
            opti.subject_to(
                optvar_battery_power_discharging
                == ca.if_else(
                    optvar_battery_power_flow < 0,
                    -optvar_battery_power_flow / (self.battery_efficiency * self.inverter_efficiency),
                    0,
                )
            )
            opti.subject_to(opti.bounded(0, optvar_battery_power_charging, self.maximum_charging_power))

            """ Energy Balance constraint for Grid , PV , Battery interaction"""

            opti.subject_to(
                optvar_power_bought_from_grid
                == ca.if_else(
                    optvar_power_thermal_delivered > 0,
                    ca.fabs(optvar_power_thermal_delivered) / cop_values
                    - optvar_power_pv
                    - optvar_battery_power_discharging,
                    ca.fabs(optvar_power_thermal_delivered) / eer_values
                    - optvar_power_pv
                    - optvar_battery_power_discharging,
                )
            )
            opti.subject_to(
                optvar_power_sold_to_grid == pv_production - optvar_battery_power_charging - optvar_power_pv
            )

        if self.flexibility_element in {"PV_only", "PV_and_Battery"}:
            opti.subject_to(opti.bounded(0, optvar_power_sold_to_grid, pv_production))
            opti.subject_to(optvar_power_bought_from_grid >= 0)
            opti.subject_to(
                optvar_power_bought_from_grid
                <= ca.if_else(
                    optvar_power_thermal_delivered > 0,
                    ca.fabs(optvar_power_thermal_delivered) / cop_values,
                    ca.fabs(optvar_power_thermal_delivered) / eer_values,
                )
            )

        """ Initial conditions """
        opti.subject_to(optvar_temperature[:, 0] == x_init)  # controlled temperature temperature
        opti.subject_to(
            optvar_disturbances == disturbance_forecast
        )  # building disturbances (solar gains / internal gains / ambient temperature)

        if self.flexibility_element in {"PV_only", "PV_and_Battery"}:
            opti.subject_to(
                optvar_power_pv_generation_forecasted == pv_production
            )  # forecasted PV generation by the generic_pv_component

        if self.flexibility_element == "PV_and_Battery":
            opti.subject_to(optvar_battery_soc[:, 0] == soc_init)  # battery state of charge

        """ Choose a concerete solver: The default linear solver used with ipopt is mumps (MUltifrontal Massively Parallel Solver).
        For a faster solution, the default solver is replaced with HSL solver 'ma27' (see 'sol_opts' > 'linear_solver').
        A free version is available for Academic Purposes ONLY. Please follow the steps provided at the end of the script to obtain,
        compile, and interface HSL solver.

        * Remark: When including the battery in the building energy system one optimization time step with { HiSim time_step = 60 sec
        and sampling rate = 1 } takes around 57 sec using the ma27 solver. The aforementioned is only for the optimization and not
        the entire household.

        This's  reduced to 0.5 sec with { HiSim time_step = 60 sec and sampling rate = 15 }
        or { HiSim time_step = 60*20 sec and sampling rate = 1 } and 'ma27' solver and even less with sampling
        rate of 20 min.

        Also it is not guarnteed that a one year simulation will work for all systems with the default solver.

        However, it is possible to do the simulation with the default solver for a building with PV installation
        ONLY.
        """

        sol_opts = {
            "ipopt": {
                "max_iter": 500,
                # "max_iter": 2000,
                "print_level": 5,
                # "print_level": 0,
                "sb": "yes",
                # "acceptable_tol": 4.0e+005,
                # "acceptable_tol": 1e-2,
                "linear_solver": "mumps",  # options: 'mumps','ma27'
                # 'linear_solver':'ma27',
                # "acceptable_obj_change_tol": 4.0e+005,
                # "acceptable_obj_change_tol": 1e-3,
            },
            "print_time": False,
        }
        opti.solver("ipopt", sol_opts)

        # numerical values of the parameter

        opti.set_value(x_init, self.state.t_m)
        opti.set_value(disturbance_forecast, disturbance_values)
        opti.set_value(cop_values, cop_sampled_array)
        opti.set_value(eer_values, eer_sampled_array)

        if self.flexibility_element in {"PV_only", "PV_and_Battery"}:
            opti.set_value(pv_production, pv_forecast_24h)

        if self.flexibility_element == "PV_and_Battery":
            opti.set_value(soc_init, self.state.soc)

        print("Starting solve", datetime.datetime.now())
        sol = opti.solve()
        # opti.debug.value

        # solution optimize resolution
        # t_m_opt=sol.value(x)
        p_th_opt = sol.value(optvar_power_thermal_delivered)
        grid_import = sol.value(optvar_power_bought_from_grid)

        # solution for actual HiSim timestep
        p_th_opt_timstep = np.repeat(p_th_opt, sampling_rate).tolist()
        grid_import_timestep = np.repeat(grid_import, sampling_rate).tolist()

        energy_efficiency_timestep = [0] * int(self.prediction_horizon)
        for i in range(int(self.prediction_horizon)):
            if p_th_opt_timstep[i] > 0:
                energy_efficiency_timestep[i] = cop_timestep[i]
            else:
                energy_efficiency_timestep[i] = eer_timestep[i]
        airconditioning_electrcitiy_consumption = [
            abs(p_th_opt_timstep[i]) / energy_efficiency_timestep[i] for i in range(int(self.prediction_horizon))
        ]

        # optimizer solution might lead to values like 1.5e-9 ---> these are replaces with zeros
        for i in range(self.prediction_horizon):
            if abs(p_th_opt_timstep[i]) < 0.1:
                p_th_opt_timstep[i] = 0
            if abs(grid_import_timestep[i]) < 0.1:
                grid_import_timestep[i] = 0
            if abs(airconditioning_electrcitiy_consumption[i]) < 0.1:
                airconditioning_electrcitiy_consumption[i] = 0

        if self.flexibility_element in {"PV_only", "PV_and_Battery"}:
            # solution optimize resolution
            pv_consumption = sol.value(optvar_power_pv)
            grid_export = sol.value(optvar_power_sold_to_grid)

            # solution for actual HiSim timestep
            pv_consumption_timestep = np.repeat(pv_consumption, sampling_rate).tolist()
            grid_export_timestep = np.repeat(grid_export, sampling_rate).tolist()

            # optimizer solution might lead to values like 1.5e-9 ---> these are replaces with zeros
            for i in range(self.prediction_horizon):
                if abs(pv_consumption_timestep[i]) < 0.1:
                    pv_consumption_timestep[i] = 0

                if abs(grid_export_timestep[i]) < 0.1:
                    grid_export_timestep[i] = 0

        if self.flexibility_element == "PV_and_Battery":
            # solution optimize resolution
            battery_to_load = sol.value(optvar_battery_power_discharging)
            pv_to_battery = sol.value(optvar_battery_power_charging)
            optvar_battery_power_flow = sol.value(optvar_battery_power_flow)
            batt_soc_actual = sol.value(optvar_battery_soc)
            batt_soc_normalized = sol.value(optvar_battery_soc) / self.maximum_storage_capacity
            if self.mpc_scheme == "optimization_once_aday_only":
                self.state.soc = batt_soc_actual[-1]
                if self.state.soc < 0.2 * self.maximum_storage_capacity:
                    self.state.soc = 0.2 * self.maximum_storage_capacity
            if self.mpc_scheme == "moving_horizon_control":
                self.state.soc = batt_soc_actual[1]

            # solution for actual HiSim timestep
            battery_to_load_timstep = np.repeat(battery_to_load, sampling_rate).tolist()
            pv_to_battery_timestep = np.repeat(pv_to_battery, sampling_rate).tolist()
            batt_soc_actual_timestep = np.repeat(batt_soc_actual, sampling_rate).tolist()
            batt_soc_normalized_timestep = np.repeat(batt_soc_normalized, sampling_rate).tolist()
            battery_power_flow_timestep = np.repeat(optvar_battery_power_flow, sampling_rate).tolist()

            # optimizer solution might lead to values like 1.5e-9 ---> these are replaces with zeros
            for i in range(self.prediction_horizon):
                if abs(battery_to_load_timstep[i]) < 0.1:
                    battery_to_load_timstep[i] = 0

                if abs(pv_to_battery_timestep[i]) < 0.1:
                    pv_to_battery_timestep[i] = 0

                if abs(batt_soc_actual_timestep[i]) < 0.1:
                    batt_soc_actual_timestep[i] = 0

                if abs(batt_soc_normalized_timestep[i]) < 0.1:
                    batt_soc_normalized_timestep[i] = 0

                if abs(battery_power_flow_timestep[i]) < 0.1:
                    battery_power_flow_timestep[i] = 0

        # calculated and can be used later if the control structure is modified to casacaded PID-
        t_m_opt_timestep = []
        t_m_init = self.state.t_m
        matrix_a_d = np.linalg.inv(identity_matrix - self.state_space_system_matrix_a)
        matrix_b_d = matrix_a_d * self.state_space_system_matrix_b
        for i in range(int(self.prediction_horizon)):
            t_m_next = (
                matrix_a_d * t_m_init
                + matrix_b_d[0, 0] * p_th_opt_timstep[i]
                + matrix_b_d[0, 1] * self.temperature_forecast_24h_1min[i]
                + matrix_b_d[0, 2] * self.temperature_forecast_24h_1min[i]
                + matrix_b_d[0, 3] * self.phi_ia_forecast_24h_1min[i]
                + matrix_b_d[0, 4] * self.phi_st_forecast_24h_1min[i]
                + matrix_b_d[0, 5] * self.phi_m_forecast_24h_1min[i]
            )
            t_m_init = t_m_next
            t_m_opt_timestep.append(t_m_next[0, 0])

        if self.flexibility_element == "basic_buidling_configuration":
            return (
                p_th_opt_timstep,
                airconditioning_electrcitiy_consumption,
                t_m_opt_timestep,
            )

        if self.flexibility_element == "PV_only":
            return (
                p_th_opt_timstep,
                airconditioning_electrcitiy_consumption,
                pv_consumption_timestep,
                grid_import_timestep,
                grid_export_timestep,
                t_m_opt_timestep,
            )

        if self.flexibility_element == "PV_and_Battery":
            return (
                p_th_opt_timstep,
                airconditioning_electrcitiy_consumption,
                pv_consumption_timestep,
                grid_import_timestep,
                grid_export_timestep,
                battery_to_load_timstep,
                pv_to_battery_timestep,
                battery_power_flow_timestep,
                batt_soc_actual_timestep,
                batt_soc_normalized_timestep,
                t_m_opt_timestep,
            )

        return None

    def cost_calculation(self, grid_export_timestep, grid_import_timestep):
        """Calculate cost of cooling consumption and revenue for buildings with renewables."""
        optimal_cost = []
        revenue = []
        for i in range(int(self.prediction_horizon)):
            costs = grid_import_timestep[i] * self.price_purchase_forecast_24h_1min[i]
            revenues = grid_export_timestep[i] * self.price_injection_forecast_24h_1min[i]
            optimal_cost.append(costs)
            revenue.append(revenues)

        return optimal_cost, revenue

    def cost_calculation_no_flexibility_element(self, airconditioning_electrcitiy_consumption):
        """Calculate cost of cooling for buildings without renewables."""
        optimal_cost = []
        for i in range(int(self.prediction_horizon)):
            costs = airconditioning_electrcitiy_consumption[i] * self.price_purchase_forecast_24h_1min[i]
            optimal_cost.append(costs)

        return optimal_cost

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:  # noqa: C901
        """Start simulation of the MPC here."""
        # t_m_old = stsv.get_input_value(self.t_m_channel)
        if self.mpcconfig.predictive:
            if (self.mpc_scheme == "optimization_once_aday_only" and timestep % self.prediction_horizon == 0) or (
                self.mpc_scheme == "moving_horizon_control"
                and timestep <= self.my_simulation_parameters.timesteps - self.prediction_horizon
            ):
                if self.my_simulation_parameters.seconds_per_timestep >= 15 * 60:
                    sampling_rate = 1
                else:
                    sampling_rate = self.sampling_rate
                scaled_horizon = int(
                    self.prediction_horizon / sampling_rate
                )  # number of points are reduced from 1440 to this value

                (
                    temperature_forecast_24h,
                    phi_ia_forecast_24h,
                    phi_st_forecast_24h,
                    phi_m_forecast_24h,
                    price_purchase_forecast_24h,
                    price_injection_forecast_24h,
                    pv_forecast_24h,
                ) = self.get_forecast_24h(timestep, sampling_rate)

                if self.flexibility_element == "basic_buidling_configuration":
                    (
                        p_th_opt_timstep,
                        airconditioning_electrcitiy_consumption,
                        t_m_opt_timestep,
                    ) = self.optimize(
                        temperature_forecast_24h,
                        phi_ia_forecast_24h,
                        phi_st_forecast_24h,
                        phi_m_forecast_24h,
                        price_purchase_forecast_24h,
                        price_injection_forecast_24h,
                        pv_forecast_24h,
                        scaled_horizon,
                    )
                    self.optimal_cost = self.cost_calculation_no_flexibility_element(
                        airconditioning_electrcitiy_consumption
                    )
                    self.revenues = [0] * self.prediction_horizon

                    self.state.cost_optimal_thermal_power = p_th_opt_timstep
                    self.air_conditioning_electricity = airconditioning_electrcitiy_consumption
                    self.cost_optimal_temperature_set_point = t_m_opt_timestep

                if self.flexibility_element == "PV_only":
                    (
                        p_th_opt_timstep,
                        airconditioning_electrcitiy_consumption,
                        pv_consumption_timestep,
                        grid_import_timestep,
                        grid_export_timestep,
                        t_m_opt_timestep,
                    ) = self.optimize(
                        temperature_forecast_24h,
                        phi_ia_forecast_24h,
                        phi_st_forecast_24h,
                        phi_m_forecast_24h,
                        price_purchase_forecast_24h,
                        price_injection_forecast_24h,
                        pv_forecast_24h,
                        scaled_horizon,
                    )
                    self.optimal_cost, self.revenues = self.cost_calculation(grid_export_timestep, grid_import_timestep)

                    self.state.cost_optimal_thermal_power = p_th_opt_timstep
                    self.air_conditioning_electricity = airconditioning_electrcitiy_consumption
                    self.pv2load = pv_consumption_timestep
                    self.electricity_from_grid = grid_import_timestep
                    self.electricity_to_grid = grid_export_timestep
                    self.cost_optimal_temperature_set_point = t_m_opt_timestep

                if self.flexibility_element == "PV_and_Battery":
                    (
                        p_th_opt_timstep,
                        airconditioning_electrcitiy_consumption,
                        pv_consumption_timestep,
                        grid_import_timestep,
                        grid_export_timestep,
                        battery_to_load_timstep,
                        pv_to_battery_timestep,
                        battery_power_flow_timestep,
                        batt_soc_actual_timestep,
                        batt_soc_normalized_timestep,
                        t_m_opt_timestep,
                    ) = self.optimize(
                        temperature_forecast_24h,
                        phi_ia_forecast_24h,
                        phi_st_forecast_24h,
                        phi_m_forecast_24h,
                        price_purchase_forecast_24h,
                        price_injection_forecast_24h,
                        pv_forecast_24h,
                        scaled_horizon,
                    )
                    self.optimal_cost, self.revenues = self.cost_calculation(grid_export_timestep, grid_import_timestep)

                    self.state.cost_optimal_thermal_power = p_th_opt_timstep
                    self.air_conditioning_electricity = airconditioning_electrcitiy_consumption
                    self.pv2load = pv_consumption_timestep
                    self.electricity_from_grid = grid_import_timestep
                    self.electricity_to_grid = grid_export_timestep
                    self.battery_to_load = battery_to_load_timstep
                    self.pv_to_battery_timestep = pv_to_battery_timestep
                    self.battery_power_flow_timestep = battery_power_flow_timestep

                    self.battery_control_state = []
                    for i in range(int(self.prediction_horizon)):
                        if battery_to_load_timstep[i] != 0:
                            state = -1  # inform the battery that energy is withdrawn
                            self.battery_control_state.append(state)
                        elif pv_to_battery_timestep[i] != 0:
                            state = 1
                            self.battery_control_state.append(state)  # inform the battery that energy is charged
                        else:
                            state = 0  # No cahrging or discharging
                            self.battery_control_state.append(state)

                    self.batt_soc_actual_timestep = batt_soc_actual_timestep
                    self.batt_soc_normalized_timestep = batt_soc_normalized_timestep
                    self.cost_optimal_temperature_set_point = t_m_opt_timestep

        if self.mpc_scheme == "optimization_once_aday_only":
            applied_optimal_solution_index = timestep % self.prediction_horizon
            self.state.t_m = self.cost_optimal_temperature_set_point[-1]
        elif self.mpc_scheme == "moving_horizon_control":
            applied_optimal_solution_index = 0
            self.state.t_m = self.cost_optimal_temperature_set_point[0]
        if (
            self.mpc_scheme == "moving_horizon_control"
            and timestep > self.my_simulation_parameters.timesteps - self.prediction_horizon
        ):
            applied_optimal_solution_index = timestep % self.prediction_horizon

        """ the air conditioner recieves electric power signal (always +ve). The binary variable operating_mode will distinguish
        heating and cooling: heating = 1 , cooling = -1 , off = 0"""
        if self.state.cost_optimal_thermal_power[applied_optimal_solution_index] > 0:
            operating_mode = 1
        elif self.state.cost_optimal_thermal_power[applied_optimal_solution_index] < 0:
            operating_mode = -1
        else:
            operating_mode = 0

        stsv.set_output_value(self.operating_mode_channel, operating_mode)
        stsv.set_output_value(
            self.p_th_mpc_channel,
            self.state.cost_optimal_thermal_power[applied_optimal_solution_index],
        )
        stsv.set_output_value(
            self.p_elec_mpc_channel,
            self.air_conditioning_electricity[applied_optimal_solution_index],
        )
        stsv.set_output_value(self.costs_channel, self.optimal_cost[applied_optimal_solution_index])
        stsv.set_output_value(self.revenues_channel, self.revenues[applied_optimal_solution_index])

        if self.flexibility_element == "basic_buidling_configuration":
            stsv.set_output_value(
                self.grid_import_channel,
                self.air_conditioning_electricity[applied_optimal_solution_index],
            )

        if self.flexibility_element in {"PV_only", "PV_and_Battery"}:
            stsv.set_output_value(
                self.pv_consumption_channel,
                self.pv2load[applied_optimal_solution_index],
            )
            stsv.set_output_value(
                self.grid_import_channel,
                self.electricity_from_grid[applied_optimal_solution_index],
            )
            stsv.set_output_value(
                self.grid_export_channel,
                self.electricity_to_grid[applied_optimal_solution_index],
            )

        if self.flexibility_element == "PV_and_Battery":
            stsv.set_output_value(
                self.battery_to_load_channel,
                self.battery_to_load[applied_optimal_solution_index],
            )
            stsv.set_output_value(
                self.pv_to_battery_channel,
                self.pv_to_battery_timestep[applied_optimal_solution_index],
            )
            stsv.set_output_value(
                self.batt_soc_actual_channel,
                self.batt_soc_actual_timestep[applied_optimal_solution_index],
            )
            stsv.set_output_value(
                self.batt_soc_normalized_channel,
                self.batt_soc_normalized_timestep[applied_optimal_solution_index],
            )
            stsv.set_output_value(
                self.battery_control_state_channel,
                self.battery_control_state[applied_optimal_solution_index],
            )
            stsv.set_output_value(
                self.battery_charging_discharging_power_channel,
                abs(self.battery_power_flow_timestep[applied_optimal_solution_index]),
            )

        """end MPC here """


"""Follow the steps below to install HSL solvers.

HSL Solvers: "HSL. A collection of Fortran codes for large scale scientific computation. http://www.hsl.rl.ac.uk/"


How to install HSL solver for Windows and interface to Casadi?

1. Submit an application for Academic Licence through: https://www.hsl.rl.ac.uk/download/coinhsl/2021.05.05/

2. Wait for an email containing a download link (No longer than one working day)

3. Download msys64 from https://www.msys2.org/

4. Follow the instruction to install mingw-w64 GCC
   pacman -S mingw-w64-ucrt-x86_64-gcc

After the download you will need pakages like OPENBLAS, LAPACK, cmake,fortran. These can be installed in Mingw64 command window with the following commands:
   pacman -S mingw-w64-x86_64-gcc-fortran
   pacman -S mingw-w64-x86_64-cmake
   pacman -S mingw-w64-x86_64-openblas
   pacman -S mingw-w64-x86_64-lapack

Depending on what is available in you machine, you may need to install other package. You will get an error if
something was missing. It is most likely easy to find a suitable solution by googling the error you got.

5. In Mingw64 command prompt, download Metis-4.0 then upack it by the following commands:
   - wget http://glaros.dtc.umn.edu/gkhome/fetch/sw/metis/OLD/metis-4.0.3.tar.gz
   - tar -xvf metis-4.0.3.tar.gz

6. Unpacked the HSL source code you obtained in step 2 and rename coinhsl-x.y.z to coinhsl

7. Move the unpacked Metis-4.03 (step 5) into the resulting folder in step 6

8. Run the command: git clone https://github.com/coin-or-tools/ThirdParty-HSL.git

9. Move the folder coinhsl to the folder ThirdParty-HSL
   - You should have a path similar to "C:----------/ThirdParty-HSL/coinhsl/metis-5.1.0"

10. Excute the following commands:
    - cd ThirdParty-HSL
    - /configure  --with-blas="-lopenblas" CXXFLAGS="-O3 -fopenmp" FCFLAGS="-O3 -fopenmp" CFLAGS="-O3 -fopenmp"
    - make
    - make install
11. After step 10, you should find a folder called ".libs". In this folder you have "libcoinhsl-0.dll"

12. You need to create a symbolic link usig the follwoign command (if needed adjust the path)

    - ln -s .libs/libcoinhsl-0.dll      .libs/libhsl.dll

13. Last to be able to use HSL with HiSim
    copy the content of the file .libs and place in  "------/.conda/envs/hisimvenv/Lib/site-packages/casadi"

Remark: After sucessfully performing the steps 1 to 13, I needed to wait for few hours until it was possible for
ipopt to see solvers. If you run into some issues after following the above step, you may need to wait for
sometime as well.

For further assistance you may write to: marwa.alfouly@tum.de
If you intend to wrok with casadi, it is helpful to check the group https://groups.google.com/g/casadi-users.

"""


"""Future Wrok.

How to enhnce receeding horizon implementaion?

Receding horizon simulation where the optimal control problem is solved at each time step is computationally
intensive. Several limitations emerged during the course of this research.
•	Performing annual simulations for economic studies of air conditioning loads with PV and battery installations
require hours.
•	Moreover, a considerable difference (could exceed one hour) in the simulation time for different regions is
observed despite the identical formulation of the optimal control problem.
•	In addition, while simulations were successfully carried out for seven locations, this is not guaranteed for
all future attempts at different sites. Even though recursive feasibility of the optimization is ensured, IPOPT
throws restoration phase failed message. After excluding possible causes like bad formulation or non-existing
derivatives, it is observed that proper scaling of the optimal control can ensure successful simulation.


Overall, proper scaling of the objective function, decision variables and other uncontrolled parameters can
facilitate the computational burden and thereby reduce simulation time. The developer of CasADi, Dr. Andersson,
highlighted the importance of scaling in optimal control even when using IPOPT that provides auto-scaling [110].
However, dynamics of heat transfer in different buildings and in different climatic conditions will be different.
For best performance, scaling factors should be simulation-specific.

For future work, it is suggested to investigate techniques for autoscaling that are customized to building HVAC
systems. It is also interesting to investigate solutions to reach 100% self-sufficiency of the air conditioning
load.

"""



"""Report Image Entry Module."""

from dataclasses import dataclass
from typing import Optional


@dataclass()
class SystemChartEntry:

    """Class for system charts entries."""

    path: str
    caption: str


class ReportImageEntry:

    """Class for report images."""

    def __init__(
        self,
        component_name: str,
        output_type: str,
        file_path: str,
        component_output_folder_path: str,
        category: Optional[str],
        output_description: Optional[str],
        unit: Optional[str],
    ) -> None:
        """Initialize the report image entry."""
        if component_name is None:
            raise ValueError("Component name was None.")
        self.component_name = component_name
        self.output_type = output_type
        self.category = category
        if output_description is None:
            raise ValueError("Component description was none from component: " + component_name)
        self.output_description = output_description
        self.unit = unit
        self.component_output_folder_path = component_output_folder_path
        self.file_path = file_path



"""Green hydrogen electrolyzer."""

# clean
import os
from typing import List, Any
import json
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from scipy.interpolate import interp1d
import numpy as np

# Import modules from HiSim
from hisim.component import SingleTimeStepValues, ComponentInput, ComponentOutput, DisplayConfig
from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters

from hisim import component as cp

__authors__ = "Franz Oldopp"
__copyright__ = "Copyright 2023, FZJ-IEK-3"
__credits__ = ["Franz Oldopp"]
__license__ = "-"
__version__ = "1.0"
__maintainer__ = "Franz Oldopp"
__status__ = "development"


@dataclass_json
@dataclass
class ElectrolyzerConfig(cp.ConfigBase):
    """Configuration of the Electrolyzer."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return Electrolyzer.get_full_classname()

    building_name: str
    name: str
    electrolyzer_type: str
    nom_load: float  # [kW]
    max_load: float  # [kW]
    nom_h2_flow_rate: float  # [m^3/h]
    faraday_eff: float
    i_cell_nom: float
    ramp_up_rate: float  # [%/s]
    ramp_down_rate: float  # [%/s]
    # H_s_h2 = 33.33 #kWh/kg

    @classmethod
    def get_default_alkaline_electrolyzer_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default Alkaline Eletrolyzer."""
        config = ElectrolyzerConfig(
            building_name=building_name,
            name="Alkaline_electrolyzer",
            electrolyzer_type="Alkaline",
            nom_load=100.0,  # [kW]
            max_load=110.0,  # [kW]
            nom_h2_flow_rate=100.0,  # [m^3/h]
            faraday_eff=1.0,
            i_cell_nom=0.3,
            ramp_up_rate=0.1,  # [%/s]
            ramp_down_rate=0.2,  # [%/s]
            # H_s_h2 = 33.33,
        )
        return config

    @staticmethod
    def read_config(electrolyzer_name):
        """Opens the according JSON-file, based on the electrolyzer_name."""

        config_file = os.path.join(utils.HISIMPATH["inputs"], "electrolyzer_manufacturer_config.json")
        with open(config_file, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)
            electrolyzer_variants = data["Electrolyzer variants"]
            if electrolyzer_name not in electrolyzer_variants:
                raise KeyError(
                    f"The electrolyzer {electrolyzer_name} could not be found in the input data. Please check the input data for electrolyzer names."
                )

            for key, values in electrolyzer_variants.items():
                if key == electrolyzer_name:
                    data_for_specific_electrolyzer = values

            return data_for_specific_electrolyzer

    @classmethod
    def config_electrolyzer(
        cls,
        electrolyzer_name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Initializes the config variables based on the JSON-file."""

        config_json = cls.read_config(electrolyzer_name)

        config = ElectrolyzerConfig(
            building_name=building_name,
            name="Electrolyzer",  # config_json.get("name", "")
            electrolyzer_type=config_json.get("electrolyzer_type"),
            nom_load=config_json.get("nom_load", 0.0),
            max_load=config_json.get("max_load", 0.0),
            nom_h2_flow_rate=config_json.get("nom_h2_flow_rate", 0.0),
            faraday_eff=config_json.get("faraday_eff", 0.0),
            i_cell_nom=config_json.get("i_cell_nom", 0.0),
            ramp_up_rate=config_json.get("ramp_up_rate", 0.0),
            ramp_down_rate=config_json.get("ramp_down_rate", 0.0),
        )
        return config


class Electrolyzer(cp.Component):
    """The Electrolszer class receives load and state inputs and calculates different outputs.

    Parameters
    ----------
    name : str
        Name of the electrolyzer variant.

    type : str
        Technology type of the electrolyzer.

    nom_load : float
        Nominal load of the electrolyzer in [kW].

    max_load : float
        Maximum load of the electrolyzer in [kW].

    nom_h2_flow_rate : float
        Hydrogen flow rate at nominal conditions in [m^3/h].

    faraday_eff : float
        Measure of the efficiency of the given electrochemical transformation.
        It is defined by given literature.

    i_cell_nom : float
        Nominal current density of the electrolyzer in [A/cm^2].

    ramp_up_rate : float
        Ramp up rate of the electrolyzer in [% of nom_load/s].

    ramp_down_rate : float
        Ramp down rate of the electrolyzer in [% of nom_load/s].

    """

    # Inputs
    LoadInput = "LoadInput"
    InputState = "InputState"

    # Outputs
    ElectrolyzerState = "ElectrolyzerState"
    CurrentLoad = "CurrentLoad"  # current load regarding the ramp-up and -down process
    TotalEnergyConsumed = "TotalEnergyConsumed"
    CurrentHydrogenFlowRate = "CurrentHydrogenFlowRate"
    CurrentOxygenFlowRate = "CurrentOxygenFlowRate"
    CurrentWaterFlowRate = "CurrentWaterFlowRate"
    TotalHydrogenProduced = "TotalHydrogenProduced"
    TotalOxygenProduced = "TotalOxygenProduced"
    TotalWaterDemand = "TotalWaterDemand"

    CurrentEfficiency = "CurrentEfficiency"
    OperatingTime = "OperatingTime"

    TotalWarmStartTime = "TotalWarmStartTime"
    WarmStartCycles = "WarmStartCycles"
    TotalColdStartTime = "TotalColdStartTime"
    ColdStartCycles = "ColdStartCycles"
    TotalRampUpTime = "TotalRampUpTime"
    TotalRampDownTime = "TotalRampDownTime"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: ElectrolyzerConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Constructs all the neccessary attributes."""
        self.electrolyzerconfig = config
        print(config)

        self.technology_type = config.electrolyzer_type
        if self.technology_type is None:
            raise ValueError("Electrolyzer type should not be None. Please choose a different electrolyzer for config.")
        self.nom_load = config.nom_load
        self.max_load = config.max_load
        self.nom_h2_flow_rate = config.nom_h2_flow_rate
        self.faraday_eff = config.faraday_eff
        self.i_cell_nom = config.i_cell_nom
        self.ramp_up_rate = config.ramp_up_rate
        self.ramp_down_rate = config.ramp_down_rate

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        # =================================================================================================================================
        # Input channels
        self.load_input: ComponentInput = self.add_input(
            self.electrolyzerconfig.name,
            Electrolyzer.LoadInput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            False,
        )

        # get the state from the controller
        self.input_state: ComponentInput = self.add_input(
            self.electrolyzerconfig.name,
            Electrolyzer.InputState,
            lt.LoadTypes.ACTIVATION,
            lt.Units.ANY,
            True,
        )
        # =================================================================================================================================
        # Output channels

        self.current_load: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.CurrentLoad,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,  # for EMS
            output_description="Current load following the input load",
        )

        self.total_energy_consumed: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.TotalEnergyConsumed,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KWH,
            output_description="Total load used for hydrogen production",
        )

        # Set total ramp-up time output
        self.total_ramp_up_time: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.TotalRampUpTime,
            lt.LoadTypes.TIME,
            lt.Units.SECONDS,
            output_description="Total ramp-up time",
        )

        # Set total ramp-down time output
        self.total_ramp_down_time: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.TotalRampDownTime,
            lt.LoadTypes.TIME,
            lt.Units.SECONDS,
            output_description="Total ramp-down time",
        )

        # Set state output
        self.electrolyzer_state: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.ElectrolyzerState,
            lt.LoadTypes.ACTIVATION,
            lt.Units.ANY,
            output_description="Current state of electrolyzer",
        )

        # current hydrogen output
        self.hydrogen_flow_rate: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.CurrentHydrogenFlowRate,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG_PER_SEC,
            output_description="Current hydrogen flow rate",
        )
        # Total hydrogen produced
        self.total_hydrogen: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.TotalHydrogenProduced,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG,
            output_description="Total hydrogen produced during simulation time",
        )
        # current oxygen output
        self.oxygen_flow_rate: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.CurrentOxygenFlowRate,
            lt.LoadTypes.OXYGEN,
            lt.Units.KG_PER_SEC,
            output_description="Current oxygen flow rate",
        )
        # Total oxygen produced
        self.total_oxygen: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.TotalOxygenProduced,
            lt.LoadTypes.OXYGEN,
            lt.Units.KG,
            output_description="Total oxygen produced during simulation time",
        )
        # current water demand
        self.water_flow_rate: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.CurrentWaterFlowRate,
            lt.LoadTypes.WATER,
            lt.Units.KG_PER_SEC,
            output_description="Current water flow rate",
        )
        # Total water demand
        self.total_water: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.TotalWaterDemand,
            lt.LoadTypes.WATER,
            lt.Units.KG,
            output_description="Total water demand during simulation time",
        )
        # Current efficiency
        self.current_efficiency_state: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.CurrentEfficiency,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description="Current efficiency based on the efficiency curve",
        )

        # Total operating time
        self.operating_time: ComponentOutput = self.add_output(
            self.electrolyzerconfig.name,
            Electrolyzer.OperatingTime,
            lt.LoadTypes.TIME,
            lt.Units.HOURS,
            output_description="Total time the electorlyzer is operating (on)",
        )
        # =================================================================================================================================
        # Transfer and storage of states

        self.current_load_state = 0.0
        self.total_ramp_up_count_state = 0.0
        self.total_ramp_down_count_state = 0.0
        self.total_warm_start_count = 0.0
        self.total_cold_start_count = 0.0
        self.total_warm_start_cycles = 0
        self.total_cold_start_cycles = 0
        self.current_warm_start_count = 0.0
        self.current_cold_start_count = 0.0
        self.total_hydrogen_produced = 0.0
        self.total_oxygen_produced = 0.0
        self.total_water_demand = 0.0
        self.total_operating_time = 0.0
        self.total_energy = 0.0

        self.current_load_previous_state = self.current_load_state
        self.total_ramp_up_count_previous_state = self.total_ramp_up_count_state
        self.total_ramp_down_count_previous_state = self.total_ramp_down_count_state
        self.total_warm_start_count_previous_state = self.total_warm_start_count
        self.total_cold_start_count_previous_state = self.total_cold_start_count
        self.total_warm_start_cycles_previous = self.total_warm_start_cycles
        self.total_cold_start_cycles_previous = self.total_cold_start_cycles
        self.current_warm_start_count_previous = self.current_warm_start_count
        self.current_cold_start_count_previous = self.current_cold_start_count
        self.total_hydrogen_produced_previous = self.total_hydrogen_produced
        self.total_oxygen_produced_previous = self.total_oxygen_produced
        self.total_water_demand_previous = self.total_water_demand
        self.total_operating_time_previous = self.total_operating_time
        self.total_energy_previous = self.total_energy

    @staticmethod
    def spec_el_stack_demand_and_polarization_data_config(
        electrolyzer_type, nominal_load, h2_flow_rate, faraday_eff, i_cell_nom
    ):
        """Polarization curve data is provided corresponding to the used electrolyzer technology.

        Following this, the auxiliary power of the system and the cell volatge is calculated,
        based on the nominal current density.
        """
        # Load data from the JSON file
        data_file = os.path.join(utils.HISIMPATH["inputs"], "electrolyzer_polarization_curve_data.json")
        with open(data_file, "r", encoding="utf-8") as file:
            data = json.load(file)

        # Check if the provided technology is valid
        if electrolyzer_type not in data:
            raise ValueError(
                f"{electrolyzer_type} is invalid technology. Supported technologies are: {', '.join(data.keys())}"
            )

        # Extract the x and y data points for the selected technology
        i_cell = data[electrolyzer_type]["i_cell"]
        u_cell = data[electrolyzer_type]["U_cell"]

        # constants
        f_constant = 96485  # C/mol
        m_h2 = 2.01588  # g/mol

        # from nom_current_density to aux_power
        spec_el_stack_demand_nom = (
            np.array(u_cell) * (2 * f_constant) / (m_h2 / 1000) / 1000 / 3600
        ) / faraday_eff  # kWh/kg
        spec_el_demand_stack = np.interp(i_cell_nom, i_cell, spec_el_stack_demand_nom)

        # calculating aux_power
        aux_power = nominal_load - (spec_el_demand_stack * h2_flow_rate)  # might needs to be set to a constant value

        # interpolarization function
        u_cell_nom = np.interp(i_cell_nom, i_cell, u_cell)  # V

        return i_cell, u_cell, i_cell_nom, u_cell_nom, aux_power

    @staticmethod
    def soec_efficiency(electrolyzer_type, current_load, max_load, state):
        """Efficiency curve data is provided corresponding to the used electrolyzer technology.

        Following this, the auxiliary power of the system and the cell volatge is calculated,
        based on the nominal current density.
        """
        # Load data from the JSON file
        data_file = os.path.join(utils.HISIMPATH["inputs"], "electrolyzer_efficiency_curve_data.json")
        with open(data_file, "r", encoding="utf-8") as file:
            data = json.load(file)

        # Check if the provided technology is valid
        if electrolyzer_type not in data:
            raise ValueError(f"Invalid technology. Supported technologies are: {', '.join(data.keys())}")

        load_perc = data[electrolyzer_type]["load_perc"]
        sys_eff = data[electrolyzer_type]["sys_eff"]

        if state == 1:
            current_load_percentage = current_load / max_load
            # Interpolation
            current_sys_eff_soec = float(np.interp(current_load_percentage, load_perc, sys_eff))

        else:
            current_sys_eff_soec = 0.0

        lhv_h2 = 33.33  # [kWh/kg]

        current_h2_production_rate = (current_sys_eff_soec * current_load) / lhv_h2  # [kg/h]

        return current_sys_eff_soec, current_h2_production_rate

    def h2_production_rate(
        self,
        i_cell_nom,
        u_cell_nom,
        nominal_load,
        i_cell,
        u_cell,
        h2_flow_rate,
        aux_power,
        current_load,
        state,
    ):
        """Hydrogen prodution rate.

        Based on the polarisation curve, the spec. electricity demand and
        the current load, the H2 production and the spec. H2 production rate
        is calculated.
        """

        nominal_power_density = i_cell_nom * u_cell_nom  # W/cm²

        h2_production_rate = np.array(i_cell) / i_cell_nom * h2_flow_rate  # kg/h

        p_cell = np.array(i_cell) * np.array(u_cell)  # W/cm²

        stack_power = p_cell / nominal_power_density * nominal_load

        # Calculates system_power from stack power
        system_power = stack_power + aux_power

        interp_function_h2_production_rate = interp1d(system_power, h2_production_rate, kind="quadratic")
        spec_h2_production_rate = h2_production_rate / system_power  # kg/kWh (proportional to the system efficiency)

        interp_function_spec_h2_production_rate = interp1d(system_power, spec_h2_production_rate, kind="quadratic")

        if state == 1:  # and current_load >= min_load:
            # Only produced hydrogen if the system is "on"
            current_h2_production_rate = float(interp_function_h2_production_rate(current_load))
            current_spec_h2_production_rate = float(interp_function_spec_h2_production_rate(current_load))
            current_eff = current_spec_h2_production_rate * 33.33  # LHV H2 33.33 kWh/kg

        else:
            # No hydrogen production of the system is in "standby" or "off"
            current_h2_production_rate = 0.0
            current_eff = 0.0

        return current_h2_production_rate, current_eff

    def oxygen_productin(self, current_h2_production_rate):
        """Oxygen production.

        Returns the produced flow rate of oxygen,
        based on the current hydrogen flow rate.
        """
        m_o2 = 31.9988
        m_h2 = 2.01588
        m_dot_o2 = (m_o2 / m_h2) * 0.5 * current_h2_production_rate  # Kurzweil (2018) - Elektrolyse von Wasser
        return m_dot_o2

    def water_demand(self, current_h2_production_rate):
        """Water demand.

        Returns the water demand flow rate,
        based on the current hydrogen flow rate.
        """
        m_h2o = 18.01528
        m_h2 = 2.01588
        m_dot_h2o = (m_h2o / m_h2) * current_h2_production_rate  # Kurzweil (2018) - Elektrolyse von Wasser
        return m_dot_h2o

    def i_save_state(self) -> None:
        """Saves the current state."""
        self.current_load_previous_state = self.current_load_state
        self.total_ramp_up_count_previous_state = self.total_ramp_up_count_state
        self.total_ramp_down_count_previous_state = self.total_ramp_down_count_state
        self.total_warm_start_count_previous_state = self.total_warm_start_count
        self.total_cold_start_count_previous_state = self.total_cold_start_count
        self.total_warm_start_cycles_previous = self.total_warm_start_cycles
        self.total_cold_start_cycles_previous = self.total_cold_start_cycles
        self.current_warm_start_count_previous = self.current_warm_start_count
        self.current_cold_start_count_previous = self.current_cold_start_count
        self.total_hydrogen_produced_previous = self.total_hydrogen_produced
        self.total_oxygen_produced_previous = self.total_oxygen_produced
        self.total_water_demand_previous = self.total_water_demand
        self.total_operating_time_previous = self.total_operating_time
        self.total_energy_previous = self.total_energy

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_restore_state(self) -> None:
        """Restores previous state."""
        self.current_load_state = self.current_load_previous_state
        self.total_ramp_up_count_state = self.total_ramp_up_count_previous_state
        self.total_ramp_down_count_state = self.total_ramp_down_count_previous_state
        self.total_warm_start_count = self.total_warm_start_count_previous_state
        self.total_cold_start_count = self.total_cold_start_count_previous_state
        self.total_warm_start_cycles = self.total_warm_start_cycles_previous
        self.total_cold_start_cycles = self.total_cold_start_cycles_previous
        self.current_warm_start_count = self.current_warm_start_count_previous
        self.current_cold_start_count = self.current_cold_start_count_previous
        self.total_hydrogen_produced = self.total_hydrogen_produced_previous
        self.total_oxygen_produced = self.total_oxygen_produced_previous
        self.total_water_demand = self.total_water_demand_previous
        self.total_operating_time = self.total_operating_time_previous
        self.total_energy = self.total_energy_previous

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""
        if force_convergence:
            return

        # get provided electricity
        p_el = stsv.get_input_value(self.load_input)

        # get state input
        state = stsv.get_input_value(self.input_state)

        # Variables
        electrolyzer_type = self.technology_type
        nominal_load = self.nom_load
        nom_h2_flow_rate = self.nom_h2_flow_rate
        faraday_eff = self.faraday_eff
        i_cell_nom = self.i_cell_nom
        ramp_up_rate = self.ramp_up_rate
        ramp_down_rate = self.ramp_down_rate
        seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep  # [s/timestep]

        # ramp up per timestep calculation
        ramp_up_per_timestep = nominal_load * ramp_up_rate * seconds_per_timestep
        total_ramp_up_per_timestep = ramp_up_per_timestep

        # ramp down per timestep calculation
        ramp_down_per_timestep = nominal_load * ramp_down_rate * seconds_per_timestep
        total_ramp_down_per_timestep = ramp_down_per_timestep

        # calculating load input
        new_load = abs(p_el - self.current_load_state)
        # pdb.set_trace()
        if state == 1:
            self.total_operating_time += seconds_per_timestep / 3600
            # pdb.set_trace()
            """the ramping process"""
            if new_load == nominal_load:
                self.current_load_state = nominal_load
                self.total_ramp_up_count_state += 0
                self.total_ramp_down_count_state += 0
                # pdb.set_trace()
            if new_load == 0:
                self.current_load_state = self.current_load_state
                self.total_ramp_up_count_state += 0
                self.total_ramp_down_count_state += 0
                # pdb.set_trace()
            # Ramping up
            if new_load >= total_ramp_up_per_timestep and self.current_load_state < p_el:
                self.total_ramp_up_count_state += seconds_per_timestep
                self.current_load_state += total_ramp_up_per_timestep
                # pdb.set_trace()
            elif self.current_load_state < p_el and new_load < total_ramp_up_per_timestep:
                percentage_ramp_up_per_timestep = new_load / total_ramp_up_per_timestep
                self.total_ramp_up_count_state += percentage_ramp_up_per_timestep * seconds_per_timestep
                # pdb.set_trace()
                if self.current_load_state == 0:
                    self.current_load_state += p_el
                    # pdb.set_trace()
                else:
                    self.current_load_state += new_load
                    # pdb.set_trace()
            # Ramping down
            elif total_ramp_down_per_timestep <= new_load and p_el < self.current_load_state:
                self.total_ramp_down_count_state += seconds_per_timestep
                self.current_load_state -= new_load
                # pdb.set_trace()
            elif p_el < self.current_load_state and new_load < total_ramp_down_per_timestep:
                percentage_ramp_down_per_timestep = new_load / total_ramp_down_per_timestep
                self.total_ramp_down_count_state += percentage_ramp_down_per_timestep * seconds_per_timestep
                self.current_load_state -= new_load
                # pdb.set_trace()
        elif state == 0:
            self.total_ramp_up_count_state += 0
            self.total_ramp_down_count_state += 0
            self.current_load_state = p_el
            # pdb.set_trace()
        elif state == -1:
            self.total_ramp_up_count_state += 0
            self.total_ramp_down_count_state += 0
            self.current_load_state = p_el
            # pdb.set_trace()

        if electrolyzer_type == "SolidOxide":
            (current_sys_eff_soec, current_h2_production_rate) = self.soec_efficiency(
                electrolyzer_type,
                self.current_load_state,
                self.max_load,
                state,
            )
        else:
            # Applying polarization curve data
            (
                i_cell,
                u_cell,
                i_cell_nom,
                u_cell_nom,
                aux_power,
            ) = self.spec_el_stack_demand_and_polarization_data_config(
                electrolyzer_type,
                nominal_load,
                nom_h2_flow_rate,
                faraday_eff,
                i_cell_nom,
            )
            # Current hydrogen prduction and specific hydrogen production rate
            (
                current_h2_production_rate,
                current_eff,
            ) = self.h2_production_rate(
                i_cell_nom,
                u_cell_nom,
                nominal_load,
                i_cell,
                u_cell,
                nom_h2_flow_rate,
                aux_power,
                self.current_load_state,
                state,
            )
        # Current oxygen and water flow rate
        current_flow_rate_oxygen = self.oxygen_productin(current_h2_production_rate)
        current_flow_rate_water = self.water_demand(current_h2_production_rate)

        # Calculating total amount of hydrogen, oxygen and water
        total_hydrogen_produced_in_timestep = current_h2_production_rate * (seconds_per_timestep / 3600)
        self.total_hydrogen_produced += total_hydrogen_produced_in_timestep
        total_oxygen_produced_in_timestep = current_flow_rate_oxygen * (seconds_per_timestep / 3600)
        self.total_oxygen_produced += total_oxygen_produced_in_timestep
        total_water_demand_in_timestep = current_flow_rate_water * (seconds_per_timestep / 3600)
        self.total_water_demand += total_water_demand_in_timestep

        self.total_energy += self.current_load_state * (seconds_per_timestep / 3600)

        # Initializing outputs
        stsv.set_output_value(self.hydrogen_flow_rate, current_h2_production_rate)
        stsv.set_output_value(self.oxygen_flow_rate, current_flow_rate_oxygen)
        stsv.set_output_value(self.water_flow_rate, current_flow_rate_water)
        stsv.set_output_value(self.electrolyzer_state, state)
        stsv.set_output_value(self.current_load, (self.current_load_state * 1000))  # Transform kW into WATT for EMS
        stsv.set_output_value(self.total_energy_consumed, self.total_energy)

        stsv.set_output_value(self.total_ramp_up_time, self.total_ramp_up_count_state)
        stsv.set_output_value(self.total_ramp_down_time, self.total_ramp_down_count_state)
        stsv.set_output_value(self.total_hydrogen, self.total_hydrogen_produced)
        stsv.set_output_value(self.total_oxygen, self.total_oxygen_produced)
        stsv.set_output_value(self.total_water, self.total_water_demand)
        stsv.set_output_value(self.operating_time, self.total_operating_time)
        if electrolyzer_type == "SolidOxide":
            stsv.set_output_value(self.current_efficiency_state, current_sys_eff_soec)
        else:
            stsv.set_output_value(self.current_efficiency_state, current_eff)

    def write_to_report(self) -> List[str]:
        """Writes a report."""
        lines = []
        for config_string in self.electrolyzerconfig.get_string_dict():
            lines.append(config_string)
        lines.append("Component Name" + str(self.component_name))
        lines.append("Total operating time during simulation: " + str(self.total_operating_time) + " [h]")
        lines.append("Total hydrogen produced during simulation: " + str(self.total_hydrogen_produced) + " [kg]")
        lines.append("Total oxygen produced during simulation: " + str(self.total_oxygen_produced) + " [kg]")
        lines.append("Total water demand during simulation: " + str(self.total_water_demand) + " [kg]")
        lines.append("Total energy consumed during simulation: " + str(self.total_energy) + " [kg]")
        lines.append("Total ramp-up time during simulation: " + str(self.total_ramp_up_count_state) + " [kg]")
        lines.append("Total ramp-down time during simulation: " + str(self.total_ramp_down_count_state) + " [kg]")
        return lines
